"""File:
"""
import traceback as _traceback

# Default indentation is four spaces.
_INDENT = '    '

def get_indent(indenter=_INDENT):
    return count_depth() * indenter

def count_depth():
    # NOTE: This doesn't work well with mutual recursion.
    stack_frames = _traceback.extract_stack()
    # stack_frames = stack_frames[:-1-parent]  # Pop myself off the stack
    depth_fn = None # stack_frames[-1][2]
    count = 0

    for filename, lineno, fn_name, text in reversed(stack_frames):
        # 'wrapper' is the default wrapped name for lru_cache, but it doesn't
        # whitelist student functions named wrapper either.
        if fn_name in ('get_indent', 'count_depth', 'wrapper'):
            continue
        if not depth_fn:
            depth_fn = fn_name
            continue
        elif fn_name == depth_fn:
            count += 1
        else:
            break
    return count



__title__ = 'campy'
__description__ = "Stanford's Graphical Libraries in Python"
__url__ = 'https://campy.sredmond.io/'
__license__ = 'MIT'

__version__ = '0.0.1.dev11'
__build__ = 0x000001
__status__ = 'Alpha'

__author__ = 'Sam Redmond'
__maintainer__ = 'Sam Redmond'
__email__ = 'sredmond@stanford.edu'
__copyright__ = '(c) 2016-2019 by Sam Redmond'
__credits__ = ['Sam Redmond', 'Alex Valderrama', 'Marty Stepp', 'Eric Roberts']

__snake__ = '\u2728 \U0001F40D \u2728'



"""Common abstract base classes for various backends.

There are three different types of backends.

- GraphicsBackendBase: Everything graphical - windows, canvas objects, labels, interactors.
- AudioBackendBase: Everything auditory - creating and playing sounds.
- ConsoleBackendBase: Everything textual - fonts and lines.

The Tk backend is a graphical backend only. The Java backend is all three.

Other implementations of `ConsoleBackendBase` have been spun out into their own
library `popup-console`.

These abstract base classes are provided for backend developers to see which
methods their backend will need to support. It is considered under development
and can change at any time.
"""
import abc

class GraphicsBackendBase(abc.ABC):
    """Interface for graphical backends."""
    # TODO(sredmond): Decorate all of these classes with @abc.abstractmethod.
    # For now, the interlaced decorators make the interface too hard to read.

    # GWindow lifecycle.
    def gwindow_constructor(self, gwindow, width, height, top_compound, visible=True): pass
    def gwindow_close(self, gwindow): pass
    def gwindow_delete(self, gwindow): pass
    def gwindow_set_exit_on_close(self, gwindow, exit_on_close): pass
    def gwindow_exit_graphics(self): pass

    # GWindow drawing.
    def gwindow_clear(self, gwindow): pass
    def gwindow_clear_canvas(self, gwindow): pass
    def gwindow_repaint(self, gwindow): pass
    def gwindow_draw(self, gwindow, gobject): pass

    # GWindow attributes.
    def gwindow_request_focus(self, gwindow): pass
    def gwindow_set_visible(self, gwindow, flag): pass
    def gwindow_set_window_title(self, gwindow, title): pass
    def gwindow_get_width(self): pass
    def gwindow_get_height(self): pass

    # GWindow geometry.
    def gwindow_add_to_region(self, gwindow, gobject, region): pass
    def gwindow_remove_from_region(self, gwindow, gobject, region): pass
    def gwindow_set_region_alignment(self, gwindow, region, align): pass

    # Shared GObject operations.
    def gobject_set_location(self, gobject, x, y): pass
    def gobject_set_filled(self, gobject, flag): pass
    def gobject_remove(self, gobject): pass
    def gobject_set_color(self, gobject, color): pass
    def gobject_set_fill_color(self, gobject, color): pass
    def gobject_send_forward(self, gobject): pass
    def gobject_send_to_front(self, gobject): pass
    def gobject_send_backward(self, gobject): pass
    def gobject_send_to_back(self, gobject): pass
    def gobject_set_size(self, gobject, width, height): pass
    def gobject_get_size(self, gobject): pass
    def gobject_get_bounds(self, gobject): pass
    def gobject_set_line_width(self, gobject, line_width): pass
    def gobject_contains(self, gobject, x, y): pass
    def gobject_set_visible(self, gobject, flag): pass
    def gobject_scale(self, gobject, sx, sy): pass
    def gobject_rotate(self, gobject, theta): pass

    # Rectangular regions
    def grect_constructor(self, grect): pass
    def groundrect_constructor(self, gobject, width, height, corner): pass
    def g3drect_constructor(self, gobject, width, height, raised): pass
    def g3drect_set_raised(self, gobject, raised): pass

    # Elliptical regions
    def goval_constructor(self, goval): pass
    def garc_constructor(self, garc): pass
    def garc_set_start_angle(self, garc, angle): pass
    def garc_set_sweep_angle(self, garc, angle): pass
    def garc_set_frame_rectangle(self, garc, x, y, width, height): pass

    # GLines
    def gline_constructor(self, gline): pass
    def gline_set_start_point(self, gline, x, y): pass
    def gline_set_end_point(self, gline, x, y): pass

    # Compounds
    def gcompound_constructor(self, gobject): pass
    def gcompound_add(self, compound, gobject): pass

    # Fonts
    def gfont_default_attributes(self): pass
    def gfont_attributes_from_system_name(self, font_name): pass
    def gfont_get_font_metrics(self, gfont): pass
    def gfont_measure_text_width(self, gfont, text): pass

    # Labels
    def glabel_constructor(self, glabel): pass
    def glabel_set_font(self, glabel, gfont): pass
    def glabel_set_label(self, glabel, text): pass

    # Polygons
    def gpolygon_constructor(self, gpolygon): pass
    def gpolygon_add_vertex(self, gpolygon, x, y): pass

    # Timers
    def gtimer_constructor(self, timer): pass
    def gtimer_delete(self, timer): pass
    def gtimer_start(self, timer): pass
    def gtimer_pause(self, timer, millis): pass
    def gtimer_stop(self, timer): pass

    # Images
    def gimage_constructor(self, gobject,  filename): pass

    def gbufferedimage_constructor(self, gobject, x, y, width, height): pass
    def gbufferedimage_fill(self, gobject, rgb): pass
    def gbufferedimage_fill_region(self, gobject, x, y, width, height, rgb): pass
    def gbufferedimage_load(self, gobject, filename): pass
    def gbufferedimage_resize(self, gobject, width, height, retain): pass
    def gbufferedimage_save(self, gobject, filename): pass
    def gbufferedimage_set_rgb(self, gobject, x, y, rgb): pass
    def gbufferedimage_update_all_pixels(self, gobject, base64): pass

    # Events
    def set_action_command(self, gobject, cmd): pass
    def get_next_event(self, mask): pass
    def wait_for_event(self, mask): pass

    def event_add_keypress_handler(self, event, handler): pass
    def event_generate_keypress(self, event): pass

    def event_add_mouse_handler(self, event, handler): pass
    def event_generate_mouse(self, event): pass

    def event_pump_one(self): pass

    # TODO(sredmond): Rename these backend events for consistency.
    def timer_pause(self, event): pass
    def timer_schedule(self, function, delay_ms): pass

    # Interactors
    def gbutton_constructor(self, gbutton): pass
    def gbutton_set_label(self, gbutton): pass
    def gbutton_set_disabled(self, gbutton): pass
    def gcheckbox_constructor(self, gcheckbox): pass
    def gcheckbox_is_selected(self, gcheckbox): pass
    def gcheckbox_set_selected(self, gcheckbox, state): pass
    def gslider_constructor(self, gslider, min, max, value): pass
    def gslider_get_value(self, gslider): pass
    def gslider_set_value(self, gslider, value): pass
    def gtextfield_constructor(self, gtextfield, num_chars): pass
    def gtextfield_get_text(self, gtextfield): pass
    def gtextfield_set_text(self, gtextfield, text): pass
    def gchooser_constructor(self, gchooser): pass
    def gchooser_add_item(self, gchooser, item): pass
    def gchooser_get_selected_item(self, gchooser): pass
    def gchooser_set_selected_item(self, gchooser, item): pass

    # Dialogs
    def gfilechooser_show_open_dialog(self, current_dir, file_filter): pass
    def gfilechooser_show_save_dialog(self, current_dir, file_filter): pass
    def goptionpane_show_confirm_dialog(self, message, title, confirm_type): pass
    def goptionpane_show_input_dialog(self, message, title): pass
    def goptionpane_show_message_dialog(self, message, title, message_type): pass
    def goptionpane_show_option_dialog(self, message, title, options, initially_selected): pass
    def goptionpane_show_text_file_dialog(self, message, title, rows, cols): pass


class ConsoleBackendBase(abc.ABC):
    """Interface for console backends."""

    @abc.abstractmethod
    def clear_console(self): pass
    @abc.abstractmethod
    def set_console_font(self, font): pass
    @abc.abstractmethod
    def set_console_size(self, console_size): pass
    @abc.abstractmethod
    def get_console_line(self): pass
    @abc.abstractmethod
    def put_console(self, line, stderr=False): pass
    @abc.abstractmethod
    def echo_console(self): pass
    @abc.abstractmethod
    def end_line_console(self): pass


class AudioBackendBase(abc.ABC):
    """Interface for audio backends."""

    # Sounds
    @abc.abstractmethod
    def create_sound(self, sound, *args): pass
    @abc.abstractmethod
    def delete_sound(self, sound, *args): pass
    @abc.abstractmethod
    def play_sound(self, sound, *args): pass

    # Notes
    @abc.abstractmethod
    def play_note(self, note, repeat): pass



"""Welcome to `campy.stanford`.

See:
    https://www.stanforddaily.com/2011/04/26/one-time-at-camp-stanford/
"""
import sys
if sys.stdout.isatty():
    print('Importing `campy.stanford`...')
    print('Import failed.')
    print('Importing `camp.stanford` instead...')
    print('☀️🌲🏕️🌲☀️')



"""Construct a backend by using the Stanford Portable Library JAR.

This module handles the initialization of and connection to the backend JAR,
as well as all interprocess communications.

This specific backend should not be directly referenced by any client applications.
"""
# TODO(sredmond): - `multiprocessing.Queue` for events to backend.
from __future__ import print_function

__PLATFORM_INCLUDED = True

import subprocess
import time

import pathlib

import collections
import re
import sys
import os
import shlex
import threading
import queue

from campy.graphics import gevents
from campy.graphics import gtypes
from campy.util import strlib

from campy.system import error

from campy.private.backends.jbe.platformat import *
from campy.private.backends.jbe.platformatter import pformat

# Constants for dialog types, taken from Java's JFileChooser
SAVE_DIALOG = 1
OPEN_DIALOG = 1

# Whether to print all pipe communications to stdout
DEBUG_PIPE = True

def boolalpha(b):
    return "true" if b else "false"

class JavaBackend:

    SOURCE_TABLE = None
    TIMER_TABLE = None
    WINDOW_TABLE = None
    BACKEND = None
    EVENT_QUEUE = None
    Q = None

    def __init__(self):
        if(JavaBackend.BACKEND == None):
            print("Initializing...")

            from campy.private.backends.jbe.jbepipe import JavaBackendPipe
            JavaBackend.BACKEND = JavaBackendPipe()

            # self.startupMain()

            # JavaBackend.Q = queue.Queue()
            # def append_output_to_queue(out, queue):
            #     for line in iter(out.readline, b''):
            #         queue.put(line)
            #     out.close()

            # t = threading.Thread(target=append_output_to_queue, args=(JavaBackend.BACKEND.stdout, JavaBackend.Q))
            # t.daemon = True # thread dies with the program
            # t.start()

            JavaBackend.EVENT_QUEUE = collections.deque()  # TODO: make me threadsafe
            JavaBackend.SOURCE_TABLE = {}
            JavaBackend.TIMER_TABLE = {}
            JavaBackend.WINDOW_TABLE = {}

### Section: GWindow
    def gwindow_add_to_region(self, gw, gobj, region):
        command = pformat(GWindow_addToRegion, id=id(gw), gobj_id=id(gobj), region=region)
        self.put_pipe(command)

    def gwindow_constructor(self, gw, width, height, top_compound, visible=True):
        JavaBackend.WINDOW_TABLE[id(gw)] = gw
        command = pformat(GWindow_constructor, id=id(gw), width=width, height=height, top_compound=id(top_compound), visible=visible)
        self.put_pipe(command)
        self.get_status()

    def gwindow_delete(self, gw):
        del JavaBackend.WINDOW_TABLE[id(gw)]
        command = pformat(GWindow_delete, id=id(gw))
        self.put_pipe(command)

    def gwindow_close(self, gw):
        command = pformat(GWindow_close, id=id(gw))
        self.put_pipe(command)

    def gwindow_request_focus(self, gw):
        command = pformat(GWindow_requestFocus, id=id(gw))
        self.put_pipe(command)

    def gwindow_set_exit_on_close(self, gw, exit_on_close):
        command = pformat(GWindow_setExitOnClose, id=id(gw), value=exit_on_close)
        self.put_pipe(command)

    def gwindow_clear(self, gw, exit_on_close):
        command = pformat(GWindow_clear, id=id(gw))
        self.put_pipe(command)

    def gwindow_clear_canvas(self, gw, exit_on_close):
        command = pformat(GWindow_clearCanvas, id=id(gw))
        self.put_pipe(command)

    def gwindow_repaint(self, gw):
        command = pformat(GWindow_repaint, id=id(gw))
        self.put_pipe(command)

    def gwindow_set_visible(self, flag, gobj = None, gw = None):
        if(gw != None):
            command = pformat(GWindow_setVisible, id=id(gw), flag=flag)
            self.put_pipe(command)
        elif(gobj != None):
            command = pformat(GObject_setVisible, id=id(gobj), flag=flag)
            self.put_pipe(command)

    def gwindow_set_window_title(self, gw, title):
        command = pformat(GWindow_setTitle, id=id(gw), title=strlib.quote_string(title))
        self.put_pipe(command)

    def gwindow_get_screen_width(self):
        command = pformat(GWindow_getScreenWidth)
        self.put_pipe(command)
        return float(self.get_result())

    def gwindow_get_screen_height(self):
        command = pformat(GWindow_getScreenHeight)
        self.put_pipe(command)
        return float(self.get_result())

    def gwindow_exit_graphics(self):
        command = pformat(GWindow_exitGraphics)
        self.put_pipe(command)

    def gwindow_draw(self, gw, gobj):
        command = pformat(GWindow_draw, id=id(gw), obj_id=id(gobj))
        self.put_pipe(command)

    def gwindow_set_region_alignment(self, gw, region, align):
        command = pformat(GWindow_setRegionAlignmnet, id=id(gw), region=region, align=align)
        self.put_pipe(command)

    def gwindow_remove_from_region(self, gw, gobj, region):
        command = pformat(GWindow_removeFromRegion, id=id(gw), obj_id=id(gobj), region=region)
        self.put_pipe(command)

### SECTION: GObject

    def gobject_set_location(self, gobj, x, y):
        command = GObject_setLocation.format(id=id(gobj), x=x, y=y)
        self.put_pipe(command)

    def gobject_set_filled(self, gobj, flag):
        command = pformat(GObject_setFilled, id=id(gobj), flag=flag)
        self.put_pipe(command)

    def gobject_remove(self, gobj):
        command = pformat(GObject_remove, id=id(gobj))
        self.put_pipe(command)

    def gobject_set_color(self, gobj, color):
        command = pformat(GObject_setColor, id=id(gobj), color=color)
        self.put_pipe(command)

    def gobject_set_fill_color(self, gobj, color):
        command = pformat(GObject_setFillColor, id=id(gobj), color=color)
        self.put_pipe(command)

    def gobject_send_forward(self, gobj):
        command = pformat(GObject_sendForward, id=id(gobj))
        self.put_pipe(command)

    def gobject_send_to_front(self, gobj):
        command = pformat(GObject_sendToFront, id=id(gobj))
        self.put_pipe(command)

    def gobject_send_backward(self, gobj):
        command = pformat(GObject_sendBackward, id=id(gobj))
        self.put_pipe(command)

    def gobject_send_to_back(self, gobj):
        command = pformat(GObject_sendToBack, id=id(gobj))
        self.put_pipe(command)

    def gobject_set_size(self, gobj, width, height):
        command = pformat(GObject_setSize, id=id(gobj), width=width, height=height)
        self.put_pipe(command)

    def gobject_get_bounds(self, gobj):
        command = pformat(GObject_getBounds, id=id(gobj))
        self.put_pipe(command)
        result = self.get_result()
        if (not result.startsWith("GRectangle(")): raise Exception(result)
        return self.scanRectangle(result)

    def gobject_set_line_width(self, gobj, line_width):
        command = pformat(GObject_setLineWidth, id=id(gobj), line_width = line_width)
        self.put_pipe(command)

    def gobject_contains(self, gobj, x, y):
        command = pformat(GObject_contains, id=id(gobj), x=x, y=y)
        self.put_pipe(command)
        return (self.get_result() == "true")

    def gobject_scale(self, gobj, sx, sy):
        command = pformat(GObject_scale, id=id(gobj), sx=sx, sy=sy)
        self.put_pipe(command)

    def gobject_rotate(self, gobj, theta):
        command = pformat(GObject_rotate, id=id(gobj), theta=theta)
        self.put_pipe(command)
### END SECTION: GObject

### SECTION: GRect
    def grect_constructor(self, gobj, width, height):
        command = pformat(GRect_constructor, id=id(gobj), width=width, height=height)
        self.put_pipe(command)
### END SECTION: GRect

### SECTION: GRoundRect
    def groundrect_constructor(self, gobj, width, height, corner):
        command = pformat(GRoundRect_constructor, id=id(gobj), width=width, height=height, corner=corner)
        self.put_pipe(command)
### END SECTION: GRoundRect

### SECTION: GCompound
    def gcompound_constructor(self, gobj):
        command = pformat(GCompound_constructor, id=id(gobj))
        self.put_pipe(command)

    def gcompound_add(self, compound, gobj):
        command = pformat(GCompound_add, compound_id=id(compound), gobj_id=id(gobj))
        self.put_pipe(command)
        self.get_status()
### END SECTION: GCompound

### SECTION: G3DRect
    def g3drect_constructor(self, gobj, width, height, raised):
        command = pformat(G3DRect_constructor, id=id(gobj), width=width, height=height, raised=raised)
        self.put_pipe(command)

    def g3drect_set_raised(self, gobj, raised):
        command = pformat(G3DRect_setRaised, id=id(gobj), raised=raised)
        self.put_pipe(command)
### END SECTION: G3DRect

### SECTION: GOval
    def goval_constructor(self, gobj, width, height):
        command = pformat(GOval_constructor, id=id(gobj), width=width, height=height)
        self.put_pipe(command)
### END SECTION: GOval

### SECTION: GArc
    def garc_constructor(self, gobj, width, height, start, sweep):
        command = pformat(GArc_constructor, id=id(gobj), width=width, height=height, start=start, sweep=sweep)
        self.put_pipe(command)

    def garc_set_start_angle(self, gobj, angle):
        command = pformat(GArc_setStartAngle, id=id(gobj), angle=angle)
        self.put_pipe(command)

    def garc_set_sweep_angle(self, gobj, angle):
        command = pformat(GArc_setSweepAngle, id=id(gobj), angle=angle)
        self.put_pipe(command)

    # TODO WTF is this method
    def garc_set_frame_rectangle(self, gobj, x, y, width, height):
        command = pformat(GArc_setFrameRectangle, id=id(gobj), x=x, y=y, width=width, height=height)
        self.put_pipe(command)
### END SECTION: GArc

### SECTION: GLine
    def gline_constructor(self, gobj, x1, y1, x2, y2):
        command = pformat(GLine_constructor, id=id(gobj), x1=x1, y1=y1, x2=x2, y2=y2)
        self.put_pipe(command)

    def gline_set_start_point(self, gobj, x, y):
        command = pformat(GLine_setStartPoint, id=id(gobj), x=x, y=y)
        self.put_pipe(command)

    def gline_set_end_point(self, gobj, x, y):
        command = pformat(GLine_setEndPoint, id=id(gobj), x=x, y=y)
        self.put_pipe(command)
### END SECTION: GLine

### SECTION: GImage
    def gimage_constructor(self, gobj,  filename):
        if(filename[0] != "/" and filename[1:3] != ":\\"):
            filename = os.getcwd() + os.sep + filename
        for i in range(len(filename)):
            if(filename[i] == "\\"):
                filename = filename[:i] + "/" + filename[i+1:]

        command = pformat(GImage_constructor, id=id(gobj), filename=filename)
        self.put_pipe(command)
        result = self.get_result()

        # TODO(sredmond): Don't return the dimension any more.
        if (not result.startswith("GDimension(")): raise Exception(result)
        return self.scanDimension(result)
### END SECTION: GImage

### SECTION: GLabel
    def glabel_constructor(self, gobj, label):
        command = pformat(GLabel_constructor, id=id(gobj), label=label)
        self.put_pipe(command)

    def glabel_set_font(self, gobj, font):
        command = pformat(GLabel_setFont, id=id(gobj), font=font)
        self.put_pipe(command)


    def glabel_set_label(self, gobj, str):
        command = pformat(GLabel_setLabel, id=id(gobj), label=strlib.quote_string(str))
        self.put_pipe(command);

    def glabel_get_font_ascent(self, gobj):
        command = pformat(GLabel_getFontAscent, id=id(gobj))
        self.put_pipe(command)
        return float(self.get_result())

    def glabel_get_font_descent(self, gobj):
        command = pformat(GLabel_getFontDescent, id=id(gobj))
        self.put_pipe(command)
        return float(self.get_result())

    def glabel_get_size(self, gobj):
        command = pformat(GLabel_getSize, id=id(gobj))
        self.put_pipe(command)
        # SO BROKEN
        return self.scanDimension(self.get_result())
### END SECTION: GLabel

### SECTION: GPolygon
    def gpolygon_constructor(self, gobj):
        command = pformat(GPolygon_constructor, id=id(gobj))
        self.put_pipe(command)

    def gpolygon_add_vertex(self, gobj, x, y):
        command = pformat(GPolygon_addVertex, id=id(gobj), x=x, y=y)
        self.put_pipe(command)
### END SECTION: GPolygon

### Section: GTimer
    def gtimer_constructor(self, timer, millis):
        JavaBackend.TIMER_TABLE[id(timer)] = timer # TODO: why?
        command = pformat(GTimer_constructor, id=id(timer), millis=millis)
        self.put_pipe(command)

    def gtimer_delete(self, timer):
        del JavaBackend.TIMER_TABLE[id(timer)]  # TODO: why?
        command = pformat(GTimer_delete, id=id(timer))
        self.put_pipe(command)

    def gtimer_start(self, timer):
        command = pformat(GTimer_start, id=id(timer))
        self.put_pipe(command)

    def gtimer_pause(self, millis):
        # TODO(sredmond): Does this method pause all active timers instead of just one? That seems wrong.
        command = pformat(GTimer_pause, millis=millis)
        self.put_pipe(command)
        self.get_status()  # TODO: wtf

    def gtimer_stop(self, timer):
        command = pformat(GTimer_stop, id=id(timer))
        self.put_pipe(command)
### End Section: GTimer

### Section: GBufferedImage
    def gbufferedimage_constructor(self, gobj, x, y, width, height):
        JavaBackend.SOURCE_TABLE[id(gobj)] = gobj
        command = pformat(GBufferedImage_constructor, id=id(gobj), x=int(x), y=int(y), width=int(width), height=int(height))
        self.put_pipe(command)

    def gbufferedimage_fill(self, gobj, rgb):
        command = pformat(GBufferedImage_fill, id=id(gobj), rgb=rgb)
        self.put_pipe(command)

    def gbufferedimage_fill_region(self, gobj, x, y, width, height, rgb):
        command = pformat(GBufferedImage_fillRegion, id=id(gobj), x=int(x), y=int(y), width=int(width), height=int(height), rgb=rgb)
        self.put_pipe(command)

    def gbufferedimage_load(self, gobj, filename):
        command = pformat(GBufferedImage_load, id=id(gobj), filename=filename)
        self.put_pipe(command)
        return self.get_result()

    def gbufferedimage_resize(self, gobj, width, height, retain):
        command = pformat(GBufferedImage_resize, id=id(gobj), width=int(width), height=int(height), retain=retain)
        self.put_pipe(command)

    def gbufferedimage_save(self, gobj, filename):
        command = pformat(GBufferedImage_save, id=id(gobj), filename=filename)
        self.put_pipe(command)
        self.get_status()  # ???

    def gbufferedimage_set_rgb(self, gobj, x, y, rgb):
        command = pformat(GBufferedImage_setRGB, id=id(gobj), x=int(x), y=int(y), rgb=rgb)
        self.put_pipe(command)

    def gbufferedimage_update_all_pixels(self, gobj, base64):
        command = pformat(GBufferedImage_updateAllPixels, id=id(gobj), base64=base64)
        self.put_pipe(command)
### End Section: GBufferedImage

### Section: Sound

    def create_sound(self, sound, filename):
        # if(filename[0] != "/" and filename[1:3] != ":\\"):
        #   filename = os.getcwd() + os.sep + filename
        # for i in range(len(filename)):
        #   if(filename[i] == "\\"):
        #       filename = filename[:i] + "/" + filename[i+1:]

        command = pformat(Sound_create, id=id(sound), filename=filename)
        self.put_pipe(command)
        # print(self.get_result())

    def delete_sound(self, sound):
        command = pformat(Sound_delete, id=id(sound))
        self.put_pipe(command)

    def play_sound(self, sound):
        command = pformat(Sound_play, id=id(sound))
        self.put_pipe(command)

### END SECTION: Sound

### SECTION: JBEConsole

    def clearConsole(self):
        command = pformat(JBEConsole_clear)
        self.put_pipe(command)

    def setConsoleFont(self, font):
        command = pformat(JBEConsole_setFont, font=font)
        self.put_pipe(command)

    def setConsoleSize(self, width, height):
        command = pformat(JBEConsole_setSize, width=width, height=height)
        self.put_pipe(command)

### END SECTION: JBEConsole

### SECTION: GInteractor

    def setActionCommand(self, gobj, cmd):
        command = pformat(GInteractor_setActionCommand, id=id(gobj), cmd=cmd)
        self.put_pipe(command)

    def getSize(self, gobj):
        command = pformat(GInteractor_getSize, id=id(gobj))
        self.put_pipe(command)
        return self.scanDimension(self.get_result())

    def gbutton_constructor(self, gobj, label):
        JavaBackend.SOURCE_TABLE[id(gobj)] = gobj
        command = pformat(GButton_constructor, id=id(gobj), label=label)
        self.put_pipe(command)

    def gcheckbox_constructor(self, gobj, label):
        JavaBackend.SOURCE_TABLE[id(gobj)] = gobj
        command = pformat(GCheckBox_constructor, id=id(gobj), label=label)
        self.put_pipe(command)

    def gcheckbox_is_selected(self, gobj):
        command = pformat(GCheckBox_isSelected, id=id(gobj))
        self.put_pipe(command)
        result = self.get_result().strip()
        return result == "true"

    def gcheckbox_set_selected(self, gobj, state):
        command = pformat(GCheckBox_setSelected, id=id(gobj), state=state)
        self.put_pipe(command)

    def gslider_constructor(self, gobj, min, max, value):
        JavaBackend.SOURCE_TABLE[id(gobj)] = gobj
        command = pformat(GSlider_constructor, id=id(gobj), min=min, max=max, value=value)
        self.put_pipe(command)

    def gslider_get_value(self, gobj):
        command = pformat(GSlider_getValue, id=id(gobj))
        self.put_pipe(command)
        return int(self.get_result())

    def gslider_set_value(self, gobj, value):
        command = pformat(GSlider_setValue, id=id(gobj), value=value)
        self.put_pipe(command)

    def createGTextField(self, gobj, num_chars):
        JavaBackend.SOURCE_TABLE[id(gobj)] = gobj
        command = pformat(GTextField_constructor, id=id(gobj), num_chars=num_chars)
        self.put_pipe(command)

    def getText(self, gobj):
        command = pformat(GTextField_getText, id=id(gobj))
        self.put_pipe(command)
        return self.get_result()

    def setText(self, gobj, str):
        command = pformat(GTextField_setText, id=id(gobj), text=strlib.quote_string(str))
        self.put_pipe(command)

    def createGChooser(self, gobj):
        JavaBackend.SOURCE_TABLE[id(gobj)] = gobj
        command = pformat(GChooser_constructor, id=id(gobj))
        self.put_pipe(command)

    def addItem(self, gobj, item):
        command = pformat(GChooser_addItem, id=id(gobj), item=strlib.quote_string(item))
        self.put_pipe(command)

    def getSelectedItem(self, gobj):
        command = pformat(GChooser_getSelectedItem, id=id(gobj))
        self.put_pipe(command)
        return self.get_result()

    def setSelectedItem(self, gobj, item):
        command = pformat(GChooser_setSelectedItem, id=id(gobj), item=strlib.quote_string(item))
        self.put_pipe(command)

### END SECTION: GInteractor

    def file_open_file_dialog(self, title, mode, path):
        # TODO: BUGFIX for trailing slashes
        command = pformat(File_openFileDialog, title=title, mode=mode, path=path)
        self.put_pipe(command)
        return self.get_result()

    def gfilechooser_show_open_dialog(self, current_dir, file_filter):
        command = pformat(GFileChooser_showOpenDialog,
            current_dir=current_dir,
            file_filter=file_filter
        )
        self.put_pipe(command)
        return self.get_result()

    def gfilechooser_show_save_dialog(self, current_dir, file_filter):
        command = pformat(GFileChooser_showOpenDialog,
            current_dir=current_dir,
            file_filter=file_filter
        )
        self.put_pipe(command)
        return self.get_result()


    def goptionpane_show_confirm_dialog(self, message, title, confirm_type):
        command = pformat(GOptionPane_showConfirmDialog,
            message=message,
            title=title,
            type=confirm_type.value
        )
        self.put_pipe(command)
        result = self.get_result()
        return int(result)

    def goptionpane_show_input_dialog(self, message, title):
        command = pformat(GOptionPane_showInputDialog,
            message=message,
            title=title
        )
        self.put_pipe(command)
        result = self.get_result()
        return strlib.url_decode(result)

    def goptionpane_show_message_dialog(self, message, title, message_type):
        command = pformat(GOptionPane_showMessageDialog,
            message=message,
            title=title,
            type=message_type.value
        )
        self.put_pipe(command)
        self.get_result()  # Wait for dialog to close

    def goptionpane_show_option_dialog(self, message, title, options, initially_selected):
        command = pformat(GOptionPane_showOptionDialog,
            message=message,
            title=title,
            options=', '.join(map(strlib.quote_string, map(strlib.url_encode, map(str, options)))),
            initial=initially_selected
        )
        self.put_pipe(command)
        result = self.get_result()
        return int(result)

    def goptionpane_show_text_file_dialog(self, message, title, rows, cols):
        command = pformat(GOptionPane_showTextFileDialog,
            message=strlib.quote_string(strlib.url_encode(message)),
            title=strlib.quote_string(strlib.url_encode(title)),
            rows=rows,
            cols=cols
        )
        self.put_pipe(command)
        self.get_result() # Wait for dialog to close

    def note_play(self, note, repeat):
        note = str(note) + boolalpha(repeat)
        command = pformat(Note_play,
            note=note
        )
        self.put_pipe(command)
        self.get_result()  # Wait for playing to be done

    ##############################
    # Section: Event Interaction
    # ----------------------------
    # This section implements interaction with the JBE console to process Events

    def getNextEvent(self, mask):
        if not JavaBackend.EVENT_QUEUE:

            command = pformat(GEvent_getNextEvent, mask=mask.value)
            self.put_pipe(command)
            self.get_result(consume_acks=True, stop_on_event=True)  # Will add to EVENT_QUEUE?
            if not JavaBackend.EVENT_QUEUE:
            # TODO: hotfix for lecture 9.1
                return gevents.GEvent()
                return None
        return JavaBackend.EVENT_QUEUE.popleft()

    def waitForEvent(self, mask):
        while not JavaBackend.EVENT_QUEUE:
            command = pformat(GEvent_waitForEvent, mask=mask.value)
            self.put_pipe(command)
            self.get_result()
        return JavaBackend.EVENT_QUEUE.popleft()

    def parseEvent(self, line):
        try:
            # TODO(sredmond): This is a broken way to parse tokens.
            # Breaks when an event parameter has a space in it.
            tokens = re.findall(r"[-\w\.\(]+", line)
            if(tokens[0] == "mousePressed("):
                return self.parseMouseEvent(tokens[1:], gevents.EventType.MOUSE_PRESSED)
            elif(tokens[0] == "mouseReleased("):
                return self.parseMouseEvent(tokens[1:], gevents.EventType.MOUSE_RELEASED)
            elif(tokens[0] == "mouseClicked("):
                return self.parseMouseEvent(tokens[1:], gevents.EventType.MOUSE_CLICKED)
            elif(tokens[0] == "mouseMoved("):
                return self.parseMouseEvent(tokens[1:], gevents.EventType.MOUSE_MOVED)
            elif(tokens[0] == "mouseDragged("):
                return self.parseMouseEvent(tokens[1:], gevents.EventType.MOUSE_DRAGGED)
            elif(tokens[0] == "keyPressed("):
                return self.parseKeyEvent(tokens[1:], gevents.EventType.KEY_PRESSED)
            elif(tokens[0] == "keyReleased("):
                return self.parseMouseEvent(tokens[1:], gevents.EventType.KEY_RELEASED)
            elif(tokens[0] == "keyTyped("):
                return self.parseKeyEvent(tokens[1:], gevents.EventType.KEY_TYPED)
            elif(tokens[0] == "actionPerformed("):
                return self.parseActionEvent(tokens[1:], gevents.EventType.ACTION_PERFORMED)
            elif(tokens[0] == "timerTicked("):
                return self.parseTimerEvent(tokens[1:], gevents.EventType.TIMER_TICKED)
            elif(tokens[0] == "windowClosed("):
                return self.parseWindowEvent(tokens[1:], gevents.EventType.WINDOW_CLOSED)
            elif(tokens[0] == "windowResized("):
                return self.parseWindowEvent(tokens[1:], gevents.EventType.RESIZED)
            elif(tokens[0] == "lastWindowClosed("):
                print("Exited normally")
                sys.exit(0)
            else:
                dummy = 1
                # ignore for now
            return gevents.GEvent()
        except Exception as inst:
            print("EXCEPTION")
            print("type:")
            print(type(inst))
            print("exception data:")
            print(inst)
            print("line:")
            print(line)
            raise
            return gevents.GEvent()

    def parseMouseEvent(self, tokens, type):
        from campy.graphics import gwindow
        id = int(tokens[0])
        tokens = tokens[1:]

        time = float(tokens[0])
        tokens = tokens[1:]

        modifiers = int(tokens[0])
        tokens = tokens[1:]

        x = float(tokens[0])
        tokens = tokens[1:]

        y = float(tokens[0])
        tokens = tokens[1:]
        print(JavaBackend.WINDOW_TABLE)

        e = gevents.GMouseEvent(type, \
                                JavaBackend.WINDOW_TABLE[id], \
                                x, \
                                y)
        # Manually set the internals of the GEvent.
        e._time = time
        e._modifiers = modifiers
        return e

    def parseKeyEvent(self, tokens, type):
        from campy.graphics import gwindow
        id = int(tokens[0])
        tokens = tokens[1:]

        time = float(tokens[0])
        tokens = tokens[1:]

        modifiers = int(tokens[0])
        tokens = tokens[1:]

        keyChar = int(tokens[0])
        tokens = tokens[1:]

        keyCode = int(tokens[0])
        tokens = tokens[1:]

        e = gevents.GKeyEvent(type, \
                                gwindow.GWindow(gwd = JavaBackend.WINDOW_TABLE[id]), \
                                keyChar, \
                                keyCode)
        # Manually set the internals of the GEvent.
        e._time = time
        e._modifiers = modifiers
        return e

    def parseTimerEvent(self, tokens, type):
        id = int(tokens[0])
        tokens = tokens[1:]

        time = float(tokens[0])
        tokens = tokens[1:]
        e = gevents.GTimerEvent(type, JavaBackend.TIMER_TABLE[id])

        # Manually set the internals of the GEvent.
        e._time = time
        return e

    def parseWindowEvent(self, tokens, type):
        from campy.graphics import gwindow
        id = tokens[0]
        tokens = tokens[1:]

        time = float(tokens[0])
        tokens = tokens[1:]

        e = gevents.GWindowEvent(type, gwindow.GWindow(JavaBackend.WINDOW_TABLE[id]))
        # Manually set the internals of the GEvent.
        e._time = time
        return e

    def parseActionEvent(self, tokens, type):
        id = int(tokens[0])
        tokens = tokens[1:]

        action = tokens[0]
        tokens = tokens[1:]

        time = float(tokens[0])
        tokens = tokens[1:]

        e = gevents.GActionEvent(type, JavaBackend.SOURCE_TABLE[id], action)

        # Manually set the internals of the GEvent.
        e._time = time
        return e

    def scanDimension(self, str):
        tokens = re.findall(r"[-:\w\.]+", str)
        #skip "GDimension"
        tokens = tokens[1:]
        width = float(tokens[0])
        tokens = tokens[1:]
        height = float(tokens[0])
        return gtypes.GDimension(width, height)

    def scanRectangle(self, str):
        tokens = re.findall(r"[-:\w\.]+", str)
        #skip "GRectangle"
        tokens = tokens[1:]
        x = float(tokens[0])
        tokens = tokens[1:]
        y = float(tokens[0])
        tokens = tokens[1:]
        width = float(tokens[0])
        tokens = tokens[1:]
        height = float(tokens[0])
        return gtypes.GRectangle(x, y, width, height)

    ##############################
    # Section: Console Interaction
    # ----------------------------
    # This section implements interaction with the JBE console for the Console class

    def get_line_console(self):
        self.put_pipe(pformat(JBEConsole_getLine))
        result = self.get_result(consume_acks=True, caller='get_line_console')
        self.echo_console(result + '\n')  # TODO: wrong for multiple inputs on one line?
        return result

    def put_console(self, line, stderr=False):
        # BUGFIX: strings that end with '\' don't print because of back-end error;
        # kludge fix by appending an "invisible" space after it
        if line.endswith('\\'):
            line += ' '
        self.put_pipe(pformat(JBEConsole_print,
            line=line,
            stderr=stderr
        ))
        self.echo_console(line, stderr)

    def echo_console(self, line, stderr=False):
        if True: # getConsoleEcho()
            needs_flush = '\n' in line
            if needs_flush:
                sys.stdout.flush()
            (sys.stdout if not stderr else sys.stderr).write(line)
            if needs_flush:
                sys.stdout.flush()
                sys.stderr.flush()

    def end_line_console(self, stderr):
        self.put_pipe(pformat(JBEConsole_println))
        self.echo_console('\n', stderr)


    ################################
    # Section: Backend Communication
    # ------------------------------
    # The following section implements utility functions to communicate with the
    # Java backend process.

    def put_pipe(self, command):
        self.BACKEND.write(command)

    def get_pipe(self):
        return self.BACKEND.read()

    def get_status(self):
        return self.BACKEND.get_status()

    def get_result(self):
        return self.BACKEND.get_result()

    # def put_pipe(self, command):
    #     # print(command)
    #     cmd = command + '\n'
    #     # out = JavaBackend.BACKEND.communicate(input=command+"\n", timeout=1)[0]
    #     if DEBUG_PIPE:
    #         print(cmd)
    #     JavaBackend.BACKEND.stdin.write(cmd)
    #     JavaBackend.BACKEND.stdin.flush()

    # def get_pipe(self):
    #     return JavaBackend.Q.get()
    #     # return JavaBackend.BACKEND.stdout.readline()

    # def get_status(self):
    #     result = self.get_result()
    #     if result != 'ok':
    #         error(result)

    # # TODO: check for whitespace returned at start or finish
    # def get_result(self, consume_acks=True, stop_on_event=False, caller=''):
    #     while True:
    #         if DEBUG_PIPE:
    #             print('getResult(): calling getPipe()...', file=sys.stderr, flush=True)
    #         line = self.get_pipe()
    #         if DEBUG_PIPE:
    #             print(line)

    #         is_result = line.startswith('result:')
    #         is_result_long = line.startswith('result_long:')
    #         is_event = line.startswith('event:')
    #         is_ack = line.startswith('result:___jbe___ack___')
    #         has_acm_exception = 'acm.util.ErrorException' in line
    #         has_exception = 'xception' in line
    #         has_error = 'Unexpected error' in line

    #         if is_result_long:
    #             # Read a long result (sent across multiple lines)
    #             result = ''
    #             next_line = self.get_pipe()
    #             while next_line != 'result_long:end':
    #                 if not line.startswith('result:___jbe___ack___'):
    #                     result += line
    #                     if DEBUG_PIPE:
    #                         print('getResult(): appended line (length so far: {})'.format(len(result)), file=sys.stderr, flush=True)
    #                 next_line = self.get_pipe()
    #             if DEBUG_PIPE:
    #                 print('getResult(): returning long strings "{}...{}" (length {})'.format(result[:10], result[-10:], len(result)), file=sys.stderr, flush=True)
    #             return result
    #         elif ((is_result or is_event) and has_acm_exception) or (not is_result and not is_event and (has_exception or has_error)):
    #             # Read an error message from the back-end
    #             if is_result:
    #                 line = line[7:]  # Prune 'result:'
    #             elif is_event:
    #                 line = line[6:]  # Prune 'event:'
    #             result = 'ERROR emitted from Stanford Java back-end process\n{}'.format(line)
    #             error(result)  # TODO: import error

    #         elif is_result:
    #             # Read a regular result
    #             if not is_ack or not consume_acks:
    #                 result = line[7:]  # Prune 'result:'
    #                 if DEBUG_PIPE:
    #                     print('getResult(): returning regular result (length {}) {}'.format(len(result), repr(result)), file=sys.stderr, flush=True)
    #                 return result.strip()
    #             else:
    #                 # Just an acknowledgement of some previous event: not a real result.
    #                 if DEBUG_PIPE:
    #                     print('getResult(): saw ACK (length {}) "{}"'.format(len(line), line), file=sys.stderr, flush=True)
    #         elif is_event:
    #             # Read a Java-originated event; enqueue it to process here.
    #             event = self.parseEvent(line[6:].strip())
    #             JavaBackend.EVENT_QUEUE.append(event)
    #             if stop_on_event or (event.event_class == gevents.EventClassType.WINDOW_EVENT and event.event_type == gevents.EventType.CONSOLE_CLOSED and caller == 'get_line_console'):
    #                 return ''
    #         else:
    #             if '\tat ' in line or '   at ' in line:
    #                 # a line from a back-end Java exception stack trace;
    #                 # shouldn't really be happening, but back end isn't perfect.
    #                 # echo it here to STDERR so Python user can see it to help diagnose the issue
    #                 print(line, file=sys.stderr, flush=True)


    # def startupMain(self):
    #     spl_location = pathlib.Path(__file__).parent / 'spl.jar'
    #     # TODO = actually communicate with the jar
    #     args = shlex.split('java -jar {}'.format(spl_location))
    #     import sys
    #     backend = subprocess.Popen(args, \
    #                                shell=False, \
    #                                stdin=subprocess.PIPE, \
    #                                stdout=subprocess.PIPE, \
    #                                stderr=sys.stdout, \
    #                                universal_newlines=True)

    #     JavaBackend.BACKEND = backend



#!/usr/bin/env python3 -tt
"""Exports a type representing 2- and 3-dimensional points.

These points inherit from tuples, so you can do anything
you could otherwise do with tuples (notably, tuple unpacking)

Usage::

    origin = Point()
    pt = Point(4, 1)
    x = origin.x  # => x = 0
    y = pt.y      # => y = 1

    origin3 = Point3()
    pt = Point3(3, 4, 5)
    x, y, z = pt
"""
import collections

Point = collections.namedtuple('Point', ['x', 'y'])
Point3 = collections.namedtuple('Point3', ['x', 'y', 'z'])

# TODO(sredmond): In Python 3.7+, use the `defaults` keyword argument to the namedtuple constructor.
Point.__new__.__defaults__= (0, 0)
Point3.__new__.__defaults__= (0, 0, 0)

# Override property documentation
# TODO(sredmond): For some reason, the property's __doc__ attribute is readonly.
# Point.x.__doc__ = "The x-coordinate of a 2D point."
# Point.y.__doc__ = "The y-coordinate of a 2D point."
# Point3.x.__doc__ = "The x-coordinate of a 3D point."
# Point3.y.__doc__ = "The y-coordinate of a 3D point."
# Point3.z.__doc__ = "The z-coordinate of a 3D point."

__all__ = ['Point', 'Point3']



"""Shuffle an immutable collection and return a copy of the collection."""
import random

def shuffle(collection, random=None):
    """Return a randomly shuffled copy of a collection.

    Note that this function's behavior depends on the type of its input.

    If collection represents an infinite iterable, this will loop indefinitely.
    """
    elements = list(collection)  # Consume all elements.
    random.shuffle(elements, random=random)
    if isinstance(collection, str):
        return ''.join(elements)
    return collection.__class__(elements)

__all__ = ['shuffle']



#!/usr/bin/env python3 -tt
"""
File: random.h
--------------
This file exports functions for generating pseudorandom numbers.
In most cases, the builtin random library should be used, but this
class offers some nice conveniences.

Note: the methods are named in consistency with the random module
"""
import random as _random
from collections import deque as _deque

"""TODO: singleton?"""
class RandomGenerator(_random.Random):
    _instance = None

    @classmethod
    def get_instance(cls):
        return cls._instance

    """ NOTE: a student shouldn't ever instantiate a RandomGenerator"""
    """ Rather, they should access one using getInstance """
    def __init__(self):
        self._fixed_bools = _deque()
        self._fixed_ints = _deque()
        self._fixed_reals = _deque()

    def randbool(self):
        return randchance(0.5)

    def randint(self, low, high):
        if self._fixed_ints:
            return self._fixed_ints.pop()
        return super().randint(low, high)

    def randchance(self, p):
        if self._fixed_bools:
            return self._fixed_bools.pop()
        return self.random() < p

    def randreal(self, low=0, high=1):
        if self._fixed_reals:
            return self._fixed_reals.pop()
        return low + (high - low) * self.random()

    def _feed_bool(self, value):
        self._fixed_bools.appendleft(value)

    def _feed_int(self, value):
        self._fixed_ints.appendleft(value)

    def _feed_real(self, value):
        self._fixed_reals.appendleft(value)

RandomGenerator._instance = RandomGenerator()



import string as _string

import campy.util.strlib as _strlib

class Platformatter(_string.Formatter):
    """Subclasess a string Formatter to support the following additional specifiers.

        !q write_quoted_string()
        !b boolalpha()  truthy values become lower-case true, else false
        !u urlencode() -> implies write_quoted

    """
    def format(self, format_string, *args, **kwargs):
        # print('format')
        # print(format_string, args, kwargs)
        return super().format(format_string, *args, **kwargs)

    def vformat(self, format_string, args, kwargs):
        # print('vformat')
        # print(format_string, args, kwargs)
        out =  super().vformat(format_string, args, kwargs)
        # print(out)
        return out

    def parse(self, format_string):
        # print('parse')
        # print(format_string)
        out = super().parse(format_string)
        # print(out)
        return out

    def get_field(self, field_name, args, kwargs):
        # print('get_field')
        # print(field_name, args, kwargs)
        return super().get_field(field_name, args, kwargs)

    def get_value(self, key, args, kwargs):
        # print('get_value')
        # print(key, args, kwargs)
        return super().get_value(key, args, kwargs)

    def check_unused_args(self, used_args, args, kwargs):
        # print('check_unused_args')
        # print(used_args, args, kwargs)
        return super().check_unused_args(used_args, args, kwargs)

    def format_field(self, value, format_spec):
        # print('format_field')
        # print(value, format_spec)
        return super().format_field(value, format_spec)

    def convert_field(self, value, conversion):
        # print('convert_field')
        # print(value, conversion)
        if conversion == 'b':
            return 'true' if value else 'false'
        elif conversion == 'q':
            return _strlib.quote_string(value)
        elif conversion == 'u':
            return _strlib.quote_string(_strlib.url_encode(value))

        return super().convert_field(value, conversion)

def _test():
    out = pformat('{0!s} {1!r} {2!b}', 4, 1, 4)
    print(out)

# Create one instance, and export its `format` method as a module-level function.

_inst = Platformatter()
pformat = _inst.format

__all__ = ['Platformatter', 'pformat']

if __name__ == '__main__':
    _test()



"""Interact with the graphics libraries via window events.

There are only two types of window events to which a caller can subscribe:

- WindowClosed: The graphical window was closed.
- WindowResized: The graphical window was resized.
"""
import campy.private.platform as _platform

# Define decorators for all of the common types of events.
def onwindowclosed():
    pass

def onwindowresized():
    pass



"""A general interval timer that fires :class:GTimerEvent`s."""
# NOTE(sredmond): Although the Stanford C++ libraries use GTimerData to manage
# reference counts, Python will automatically garbage collect correctly for us.
# Overriding the __del__ method allows us to offer Python some instruction for
# when a GTimer is cleaned up, but __del__ isn't guaranteed to be called
# immediately or at all, so we risk stranding memory.
import campy.private.platform as _platform


class GTimer:
    """An interval timer that generates :class:GTimerEvent`s at a fixed rate."""

    def __init__(self, milliseconds):
        """Create a :class:`GTimer` that can generate :class:`GTimerEvent`s.

        Once started, the timer will generate a :class:`GTimerEvent` each time
        the given number of milliseconds has elapsed.

        Note that no events are generated until the client calls :meth:`start()`
        on the timer.

        For more, see the documentation for :class:`GTimerEvent`.

        :param milliseconds: The timer interval (in milliseconds).
        """
        self.milliseconds = milliseconds
        _platform.Platform().gtimer_constructor(self)

    def __del__(self):
        """Delete the backend resources associated with this :class:`GTimer`.

        This method is called by Python only when there are no references to
        this object. However, it is not guaranteed to be called immediately or
        even eventually (depending on the Python implementation). Thus, it's
        best to think of this as a friendly suggestion to the Python
        interpreter that cleaning up a :class:`GTimer` involves informing the
        backend that it should free up any resources associated with the timer
        as well.
        """
        # TODO(sredmond): Investigate race condition when GTimer is deleted
        # after the platform is deleted.
        _platform.Platform().gtimer_delete(self)

    def start(self):
        """Start this timer.

        A timer will generate :class:`GTimerEvent`s until it is stopped.

        To achieve the effect of a one-shot timer, the best approach is to
        call :meth:`stop()` on this timer after processing one event.
        """
        _platform.Platform().gtimer_start(self)

    def pause(self, milliseconds):
        """Pause this timer for some time before automatically resuming.

        :param milliseconds: How long to pause this timer for (in milliseconds).
        """
        _platform.Platform().gtimer_pause(self, millis)

    def stop(self):
        """Stop this timer from generating events until it is restarted."""
        _platform.Platform().gtimer_stop(self)

    def __repr__(self):
        return 'GTimer(ms={})'.format(self.milliseconds)



"""Exported types for representing points, dimensions, and rectangles.

A :class:`GPoint` represents an (x, y) pair and defaults to (0, 0).

A :class:`GDimension` represents a (width, height) pair and defaults to (0, 0).

A :class:`GRectangle` represents an (x, y, width, height) quadruplet and
defaults to (0, 0, 0, 0).

Named attributes are accessible for each of these types::

    point = GPoint(4, 1)
    print(point.x, point.y)  # => 4, 1

    size = GDimension(41, 574)
    print(size.width, size.height)  # => 41, 574

    rect = GRectangle(4, 1, 41, 574)
    print(rect.x, rect.y, rect.width, rect.height)  # => 4, 1, 41, 574

Each of these types can also be unpacked like a tuple.

    point = GPoint(4, 1)
    x, y = point

    size = GDimension(41, 574)
    width, height = size

    rect = GRectangle(4, 1, 41, 574)
    x, y, width, height = rect

Lastly, each of these types is immutable.
"""
# NOTE: There's a lot of hacking around with the internals of namedtuple in
# this module. This level of hacking is not encouraged, and might be removed
# in the future, replacing these types with custom-style namedtuples.
# Here be dragons.
# TODO(sredmond): Redocument these types so that Sphinx's autodoc sees them.
# TODO(sredmond): Reconsider defaulting all arguments to 0.0 instead of 0.
import collections
import operator


# An (x, y) pair defaulting to (0, 0).
GPoint = collections.namedtuple('GPoint', ('x', 'y'))
GPoint.__new__.__defaults__ = (0, 0)


# A (width, height) pair defaulting to (0, 0).
GDimension = collections.namedtuple('GDimension', ('width', 'height'))
GDimension.__new__.__defaults__ = (0, 0)
GDimension.empty = lambda dim: dim.width <= 0 or dim.height <= 0
GDimension.empty.__name__ = 'empty'


# An (x, y, width, height) quadruplet defaulting to (0, 0, 0, 0)
GRectangle = collections.namedtuple('GRectangle', ('x', 'y', 'width', 'height'))
GRectangle.__new__.__defaults__ = (0, 0, 0, 0)
GRectangle.empty = lambda rect: rect.width <= 0 or rect.height <= 0
GRectangle.empty.__name__ = 'empty'
# NOTE(sredmond): Since we consider the edge to be contained, "empty" GRectangles can still contain points.
GRectangle.__contains__ = lambda rect, point: 0 <= point[0] - rect.x <= rect.width and 0 <= point[1] - rect.y <= rect.height
GRectangle.__contains__.__name__ = '__contains__'


# Override docstrings. This is perhaps the messiest part of this file.
GPoint.__doc__ = 'Graphical representation of an (x, y) point defaulting to (0, 0).'
GPoint.__new__.__doc__ = """Create a :class:`GPoint`.

If parameters are not supplied, all arguments default to 0.

All arguments have units of pixels.

:param x: The x-coordinate (in pixels).
:param y: The y-coordinate (in pixels).
"""
GPoint.x = property(operator.itemgetter(0), doc='The x-coordinate (in pixels).')
GPoint.y = property(operator.itemgetter(1), doc='The y-coordinate (in pixels).')


GDimension.__doc__ = 'Graphical representation of a size (width, height) defaulting to (0, 0).'
GDimension.__new__.__doc__ = """Create a :class:`GDimension`.

If parameters are not supplied, all arguments default to 0.

All arguments have units of pixels.

:param width: The width (in pixels).
:param height: The height (in pixels).
"""
GDimension.width = property(operator.itemgetter(0), doc='The width (in pixels).')
GDimension.height = property(operator.itemgetter(1), doc='The height (in pixels).')
GDimension.empty.__doc__ = 'Return whether this rectangle is empty.'


GRectangle.__doc__ = """Graphical rectangle with upper-left corner (x, y) and size (width, height).

A :class:`GRectangle` is often used to represent the bounding box of a graphical object.
"""
GRectangle.__new__.__doc__ = """Create a :class:`GRectangle`.

The x- and y-coordinates refer to the upper-left corner of this
:class:`GRectangle`.

If parameters are not supplied, all arguments default to 0.

All arguments have units of pixels.

:param x: The x-coordinate of the upper left corner (in pixels).
:param y: The y-coordinate of the upper left corner (in pixels).
:param width: The width of this rectangle (in pixels).
:param height: The height of this rectangle (in pixels).
"""
GRectangle.x = property(operator.itemgetter(0), doc='The x-coordinate of the upper left corner (in pixels).')
GRectangle.y = property(operator.itemgetter(1), doc='The y-coordinate of the upper left corner (in pixels).')
GRectangle.width = property(operator.itemgetter(2), doc='The width of this rectangle (in pixels).')
GRectangle.height = property(operator.itemgetter(3), doc='The height of this rectangle (in pixels).')
GRectangle.empty.__doc__ = 'Return whether this rectangle is empty.'
GRectangle.__contains__.__doc__ = """Return whether this rectangle contains a point.

The edges of the rectangle are considered to be contained by the rectangle.
For example, the point (4, 1) is contained in the rectangle (4, 1, 41, 574).

The point must be unpackable into two elements, which means that you can use a
:class:`GPoint` or a two-element tuple::

    rect = GRectangle(4, 1, 41, 574)
    point = GPoint(4, 1)
    if point in rect:  # Using a GPoint.
        print('GPoint is contained.')
    if (5, 3) in rect:  # Using a tuple.
        print('tuple is contained')

:param point: An (x, y) point to check for containment.
:return: Whether this rectangle contains the given point.
"""

__all__ = ('GPoint', 'GDimension', 'GRectangle')



# TODO(sredmond): Implement hash.

class Vertex():
    """Canonical Vertex (Node) structure used by :class:`BasicGraph`.

    """
    # TODO(sredmond): Investigate whether VertexGen in the C++ library is meaningful.
    def __init__(self, name):
        self.name = name
        self.reset_data()

    ## TODO Copy constructor

    def reset_data(self):
        self.arcs = set()  # {Edge}
        self.visited = False
        self.previous = None
        self.cost = 0.0
        self._color = 0  # UNCOLORED

    @property
    def color(self):
        """Return the color of this vertex."""
        return self._m_color

    @color.setter
    def color(self, c):
        self._color = c
        # self.notify_observers()  ## TODO Add observables

    # Convenient aliases
    @property
    def edges(self):
        return self.arcs

    @property
    def weight(self):
        return self.cost

    def __repr__(self):
        return "Vertex(name={}, cost={}, visited={}, previous={}, neighbors={{}})".format(
            self.name,
            self.cost,
            self.visited,
            self.previous.name if self.previous else 'None',
            ', '.join(edge.finish.name for edge in self.edges)
        )

    ## TODO equality override

# Node = Vertex  # Should we allow class-level aliasing like Marty?

class Edge():
    def __init__(self, start=None, finish=None, cost=0.0):
        self.start = start
        self.finish = finish
        self.cost = cost
        self.reset_data()

    def reset_data(self):
        self.visited = False

    # Convenient aliases
    @property
    def end(self):
        return self.finish

    @end.setter
    def end(self, value):
        self.finish = value

    @end.deleter
    def end(self):
        del self.end

    @property
    def weight(self):
        return self.cost

    def __repr__(self):
        return "Edge(start={}, finish={}, cost={}, visited={})".format(
            self.start.name if self.start else 'None',
            self.finish.name if self.finish else 'None',
            self.cost,
            self.visited
        )

# Arc = Edge  # Should we alias this too?

class BasicGraph():
    def __init__(self):
        self._reset_enabled = True



import time as _time

class Timer():
    def __init__(self, autostart=False):
        """Constructs a new Timer.

        If autostart is true, immediately starts the timer.
        """
        self._start_millis = 0
        self._stop_millis = 0
        self._started = False
        if autostart:
            self.start()

    def elapsed(self):
        return self._stop_millis - self._start_millis

    def is_started(self):
        return self._started

    def start(self):
        self._start_millis = self.current_time_millis()
        self._started = True

    def stop(self):
        self._stop_millis = self.current_time_millis()
        if not self.is_started():
            # error("Timer is not started!")
            self._start_millis = self._stop_millis
        self._started = False
        return self.elapsed()

    @classmethod
    def current_time_millis(cls):
        return int(round(_time.time() * 1000))



"""Show graphical dialog boxes that ask the user to select a file.

These dialog boxes come in two flavors:

- Open: Select a file to open.
- Save: Select a file to save.

In each case, the filename is returned, or None if the user closed the dialog.
"""
# TODO(sredmond): Reconsider having the empty string returned as a sentinel.
# TODO(sredmond): Add support for additional parameters to the dialogs.
# e.g. defaultextension, title)
import campy.private.platform as _platform

import os


# The current user's home directory, used as the default initial directory.
HOME = os.path.expanduser('~')


def show_open_dialog(directory=HOME):
    """Show the user a dialog asking for a file to open.

    :param directory: The initial directory displayed in the dialog.
    :type directory: pathlib.Path or str
    :return: The filename of the user-selected file, or None if the dialog was cancelled.
    """
    return _platform.Platform().gfilechooser_show_open_dialog(str(directory), file_filter='')


def show_save_dialog(directory=HOME):
    """Show the user a dialog asking for a file to save.

    :param directory: The initial directory displayed in the dialog.
    :type directory: pathlib.Path or str
    :return: The filename of the user-selected file, or None if the dialog was cancelled.
    """
    return _platform.Platform().gfilechooser_show_save_dialog(str(directory), file_filter='')



"""Provide a :class:`GWindow` that supports drawing graphical objects on screen."""
from collections import namedtuple as _nt
import enum as _enum
import math

import campy.graphics.gtypes as _gtypes
import campy.graphics.gobjects as _gobjects
import campy.graphics.gtypes as _gtypes
import campy.private.platform as _platform
import campy.graphics.gcolor as _gcolor


@_enum.unique
class Alignment(_enum.Enum):
    """Horizontal alignment within a region."""
    LEFT = 0
    CENTER = 1
    RIGHT = 2


@_enum.unique
class Region(_enum.Enum):
    """Regions of a :class:`GWindow`."""
    CENTER = 0
    EAST = 1
    NORTH = 2
    SOUTH = 3
    WEST = 4


@_enum.unique
class CloseOperation(_enum.Enum):
    """Actions to take upon closure of a GWindow.

    Currently unused.
    """
    DO_NOTHING = 0
    HIDE = 1
    DISPOSE = 2
    EXIT = 3


# TODO(sredmond): There used to be a lot of fluff around copying a GWindow.
# It seems like that was all misguided - Python should handle the creation
# of these objects. However, keep an eye out for unexpected copying errors.
class GWindow:
    """A Graphics Window that supports simple graphics.

    Each GWindow consists of two layers. The background layer
    provides a surface for drawing static pictures that involve no animation.
    Graphical objects drawn in the background layer are persistent and do
    not require the client to update the contents of the window.  The
    foreground layer contains graphical objects that are redrawn as necessary.

    The GWindow class includes several methods that draw
    lines, rectangles, and ovals on the background layer without making
    use of the facilities of the gobjects.h interface.  For
    example, the following program draws a diamond, rectangle, and oval
    at the center of the window::

        gw = gwindow.GWindow
        print("This program draws a diamond, rectangle, and oval.")
        width = gw.getWidth()
        height = gw.getHeight()
        gw.drawLine(0, height / 2, width / 2, 0);
        gw.drawLine(width / 2, 0, width, height / 2);
        gw.drawLine(width, height / 2, width / 2, height);
        gw.drawLine(width / 2, height, 0, height / 2);
        gw.setColor("BLUE");
        gw.fillRect(width / 4, height / 4, width / 2, height / 2);
        gw.setColor("GRAY");
        gw.fillOval(width / 4, height / 4, width / 2, height / 2);
    """
    # The default width (in pixels) for a new GWindow.
    DEFAULT_WIDTH = 500

    # The default height (in pixels) for a new GWindow.
    DEFAULT_HEIGHT = 500

    # TODO(sredmond): Add a default color once the GColor library is fixed!
    def __init__(self, width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, visible=True, title="", color=None, top=None):
        """Create a new GWindow. Optionally, supply a specific width and height.

        Initial visibility, window title, marker color, and top compound can also
        be specified, but will default to reasonable values.

        To create a new GWindow of a default size::

            window = GWindow()

        To create a new GWindow with a specific size::

            window = GWindow(width=1280, height=800)

        To create a new GWindow with the default size but a specific title and marker color::

            window = GWindow(title="My Window", color=GColor.RED)

        :param width: The initial width of the GWindow.
        :param height: The initial height of the GWindow.
        :param visible: The default visibility of the GWindow.
        :param title: The displayed title on the GWindow.
        :param color: The marker color used to draw GObjects.
        :param top: The topmost GCompound in the GWindow.
        """
        self._width = width
        self._height = height
        self._visible = visible
        self._title = title
        # TODO(sredmond): Propagate the title to the backend constructor.

        if not color:
            color = _gcolor.GColor.BLACK
        self._color = color

        if not top:
            top = _gobjects.GCompound()
        self._top = top

        # TODO(sredmond): Isn't it a little silly to pass this object along with its attributes?
        _platform.Platform().gwindow_constructor(self, self._width, self._height, self._top)

    def close(self):
        """Close this GWindow."""
        # TODO(sredmond): Use the CloseOperation setting to determine how aggresive to close things down.
        _platform.Platform().gwindow_close(self)
        _platform.Platform().gwindow_delete(self)

    @property
    def width(self):
        """Get this GWindow's width (in pixels)."""
        # TODO(sredmond): If the actual GWindow's width changes (i.e. from a resize operation),
        # this value will be incorrect.
        return self._width

    @property
    def height(self):
        """Get this GWindow's height (in pixels)."""
        # TODO(sredmond): If the actual GWindow's height changes (i.e. from a resize operation),
        # this value will be incorrect.
        return self._height

    @property
    def visible(self):
        """Get or set this GWindow's visibility.

        TODO(sredmond): Document what, exactly, a GWindow's visibility means.

        Usage::

            window = GWindow()
            if not window.visible:
                window.visible = True
        """
        return self._visible

    @visible.setter
    def visible(self, flag):
        self._visible = flag
        _platform.Platform().gwindow_set_visible(flag, gw=self)

    @property
    def title(self):
        """Get or set this GWindow's title.

        Changing the title modifies the text in the top bar of the GWindow.

        Usage::

            window = GWindow()
            window.title = "My Title"
            print(window.title)
        """
        return self._title

    @title.setter
    def title(self, title):
        self._title = title
        _platform.Platform().gwindow_set_window_title(self, title)

    @property
    def color(self):
        """Get or set the marker color used for drawing.

        When accessing the color, retrieve a GColor object.

        TODO(sredmond): Document what sorts of colors can be supplied.
        """
        return self._color

    @color.setter
    def color(self, color):
        self._color = color
        # TODO(sredmond): Canonicalize this input color into a GColor.

    def draw_line(self, x0, y0, x1, y1):
        """Draw a line between the given coordinates.

        The line's color is this GWindow's current marker color.

        To draw a line from (0, 0) to (4, 1)::

            window = GWindow()
            window.draw_line(0, 0, 4, 1)

        To draw a line between two :class:`GPoint`s::

            window = GWindow()
            window.draw_line(*p0, *p1)

        :param x0: The x-coordinate of the starting point.
        :param y0: The y-coordinate of the starting point.
        :param x1: The x-coordinate of the ending point.
        :param y1: The y-coordinate of the ending point.
        """
        # TODO(sredmond): Multiple splat unpackings weren't supported until a recent version of Python.
        line = _gobjects.GLine(x0, y0, x1, y1)
        line.color = self.color
        self.draw(line)

    def draw_polar_line(self, x, y, r, theta):
        """Draw a line from an initial point with a given length and polar direction.

        The angle is measured in degrees counterclockwise from the positive x-axis.

        This function also returns the ending point of the constructed line.

        To draw a line of length 3 from the point (100, 100) at an angle of 60 degrees::

            window = GWindow()
            window.draw_polar_line(100, 100, 3, 60)

        :param x: The x-coordinate of the starting point.
        :param y: The y-coordinate of the starting point.
        :param r: The length of the line to draw.
        :param theta: The angle (measured from the positive x-axis) of the new line.
        :returns: The endpoint of the line.
        """
        # TODO(sredmond): Consider alternate ways to specify coordinates with a GPoint.
        x_end = x + r * math.cos(math.radians(theta))
        y_end = y - r * math.sin(math.radians(theta))  # Subtraction, since y decreases upwards.
        self.draw_line(x, y, x_end, y_end)
        return _gtypes.GPoint(x_end, y_end)

    def draw_oval(self, x, y, width, height):
        """Draw the outline of an oval.

        The oval's upper-left corner has coordinates given by (x, y), and size
        given by (width, height). The corner of an oval is the corner of the oval's
        bounding box.

        The color of the oval's outline will be this GWindow's current marker color.

        To draw an oval with an upper-left corner at (0, 0), a width of 300 pixels,
        and a height of 200 pixels::

            window = GWindow()
            window.draw_oval(0, 0, 300, 200)

        To draw a circle centered at (410, 410) with radius 160::

            window = GWindow()
            window.draw_oval(410 - 160, 410 - 160, 160 * 2, 160 * 2)

        :param x: The x-coordinate of the upper-left corner of the oval's bounding box.
        :param y: The y-coordinate of the upper-left corner of the oval's bounding box.
        :param width: The width of the oval in pixels.
        :param height: The height of the oval in pixels.
        """
        # TODO(sredmond): Find a way to supply a GRectangle's bounds instead.
        # TODO(sredmond): It's a little awkward to reverse the order of the arguments here.
        oval = _gobjects.GOval(width, height, x=x, y=y)
        # oval.color = self.color  # TODO(sredmond): Uncomment me! Just commented out for a brief moment of testing.
        self.draw(oval)

    def fill_oval(self, x, y, width, height):
        """Draw a filled oval.

        The oval's upper-left corner has coordinates given by (x, y), and size
        given by (width, height). The corner of an oval is the corner of the oval's
        bounding box.

        The color of the oval will be this GWindow's current marker color.

        To draw a filled oval with an upper-left corner at (0, 0), a width of 300
        pixels, and a height of 200 pixels::

            window = GWindow()
            window.fill_oval(0, 0, 300, 200)

        To draw a filled circle centered at (410, 410) with radius 160::

            window = GWindow()
            window.fill_oval(410 - 160, 410 - 160, 160 * 2, 160 * 2)

        :param x: The x-coordinate of the upper-left corner of the oval's bounding box.
        :param y: The y-coordinate of the upper-left corner of the oval's bounding box.
        :param width: The width of the oval in pixels.
        :param height: The height of the oval in pixels.
        """
        # TODO(sredmond): Find a way to supply a GRectangle's bounds instead.
        # TODO(sredmond): It's a little awkward to reverse the order of the arguments here.
        oval = _gobjects.GOval(width, height, x=x, y=y)
        oval.color = self.color
        # TODO(sredmond): Possibly, fall back to a black outline when the oval is filled.
        # oval.fill_color = self.color # TODO(sredmond): Uncomment me! Just commented out for a brief moment of testing.
        # oval.filled = True  # TODO(sredmond): Uncomment me! Just commented out for a brief moment of testing.
        self.draw(oval)

    def draw_rect(self, x, y, width, height):
        """Draw the outline of a rectangle.

        The rectangle's upper-left corner has coordinates given by (x, y) and size
        given by (width, height).

        The color of the rectangle's outline will be this GWindow's current color.

        To draw a rectangle with an upper-left corner at (0, 0), a width of 300 pixels,
        and a height of 200 pixels::

            window = GWindow()
            window.draw_rectangle(0, 0, 300, 200)

        :param x: The x-coordinate of the upper-left corner of the rectangle's bounding box.
        :param y: The y-coordinate of the upper-left corner of the rectangle's bounding box.
        :param width: The width of the rectangle in pixels.
        :param height: The height of the rectangle in pixels.
        """
        # TODO(sredmond): Find a way to supply a GRectangle's bounds instead.
        # TODO(sredmond): It's a little awkward to reverse the order of the arguments here.
        rect = _gobjects.GRect(width, height, x=x, y=y)
        rect.color = self.color
        self.draw(rect)

    def fill_rect(self, x, y, width, height):
        """Draw a filled rectangle.

        The rectangle's upper-left corner has coordinates given by (x, y) and size
        given by (width, height).

        The color of the rectangle's outline will be this GWindow's current color.

        To draw a rectangle with an upper-left corner at (0, 0), a width of 300 pixels,
        and a height of 200 pixels::

            window = GWindow()
            window.draw_rectangle(0, 0, 300, 200)

        :param x: The x-coordinate of the upper-left corner of the rectangle's bounding box.
        :param y: The y-coordinate of the upper-left corner of the rectangle's bounding box.
        :param width: The width of the rectangle in pixels.
        :param height: The height of the rectangle in pixels.
        """
        # TODO(sredmond): It's a little awkward to reverse the order of the arguments here.
        rect = _gobjects.GRect(width, height, x=x, y=y)
        rect.color = self.color
        rect.fill_color = self.color
        rect.filled = True
        self.draw(rect)

    def draw(self, gobj, x=None, y=None):
        """Draw a GObject on this GWindow's background layer.

        The background layer is for static drawings that cannot be modified once drawn.

        If both x and y are supplied, the object is moved to that location before drawing.

        :param gobj: The GObject to draw to this GWindow's background layer.
        :param x: The x-coordinate at which to draw the GObject.
        :param y: The y-coordinate at which to draw the GObject.
        """
        if x is not None and y is not None:
            gobj.location = x, y
        _platform.Platform().gwindow_draw(self, gobj)

    def clear(self):
        """Clear all content from this GWindow.

        Both the background and foreground layers are cleared with this function.
        """
        _platform.Platform().gwindow_clear(self)  # Remove from background layer.
        self._top.clear()  # Remove from foreground layer.

    def add(self, gobj, x=None, y=None):
        """Add a :class:`GObject` to the foreground layer of this :class:`GWindow`.

        If both x and y are supplied, the object is moved to that location before adding.

        To add a GOval to the window::

            window = GWindow()
            oval = GOval(0, 0, 100, 100)
            window.add(oval)

        :param gobj: The GObject to draw to this GWindow's foreground layer.
        :param x: The x-coordinate at which to draw the GObject.
        :param y: The y-coordinate at which to draw the GObject.
        """
        # TODO(sredmond): Handle the case where location is a float.
        if x is not None and y is not None:
            gobj.location = x, y
        self._top.add(gobj)

    def __iadd__(self, gobj):
        """Implement ``self += gobj``.

        Add a GObject to the foreground layer of this :class:`GWindow`.
        """
        self.add(gobj)

    def remove(self, gobj):
        """Remove a :class:`GObject` from this :class:`GWindow`.

        If the GObject was in the foreground layer of the window, return True.
        Otherwise, return False, but otherwise do not any failure.

        :param gobj: The GObject to remove from this GWindow.
        :returns: Whether the GObject was previously contained in this GWindow.
        """
        return self._top.remove(gobj)

    def __isub__(self, obj):
        """Implement ``self -= gobj``.

        Remove a GObject from the foreground layer of this :class:`GWindow`.
        """
        # Ignore the return value.
        self.remove(gobj)

    def add_to_region(self, gobj, region):
        """Add an interactor to the control strip in a given region.

        The interactor could also be a GLabel.

        The region argument must be some region from the :class:`Region` enum.

        To add a button to the SOUTH region::

            window = GWindow()
            button = GButton("Click me!")
            window.add_to_region(button, Region.SOUTH)

        :param gobj: The interactor to add to a region.
        :param region: The region to which the interactor will be added.
        :type region: Region
        """
        # TODO(sredmond): Either here, or at the platform level, convert into the region as a string.
        _platform.Platform().gwindow_add_to_region(self, gobj, region)

    def remove_from_region(self, gobj, region):
        """Remove an interactor from the control strip in a given region.

        The region argument must be some region from the :class:`Region` enum.

        :param gobj: The interactor to remove from a region.
        :param region: The region from which the interactor will be removed.
        :type region: Region
        """
        # TODO(sredmond): Either here, or at the platform level, convert into the region as a string.
        _platform.Platform().gwindow_remove_from_region(self, gobj, region)

    def set_region_alignment(self, region, align):
        """Set an alignment for a given region.

        Both the region and alignment arguments must be from the :class:`Region`
        and :class:`Alignment` enums respectively.

        To CENTER the content in the SOUTH region::

            window = GWindow()
            window.set_region_alignment(Region.SOUTH, Alignment.CENTER)

        :param region: The region in which to set alignment.
        :type region: Region
        :param align: The alignment to set in the region.
        :type align: Alignment
        """
        # TODO(sredmond): Either here, or at the platform level, convert into the region and alignment as a string.
        _platform.Platform().gwindow_set_region_alignment(self, region, align)

    def get_object_at(self, x, y):
        """Return the topmost GObject containing the point (x, y) or None.

        If no GObject contains the point (x, y), then return None.

        This only searches through the foreground layer.

        To search for a GObject at the point (41, 106)::

            window = GWindow()
            gobj = window.get_object_at(41, 106)
            if gobj is not None:
                print('We found an object!')
            else:
                print('No object found.')

        :param x: The x-coordinate of the point to examine.
        :param y: The y-coordinate of the point to examine.
        :returns: The topmost GObject containing the given point, or None if no such object was found.
        """
        # TODO(sredmond): This is currently implemented as an inefficient linear scan.
        # Consider optimizing the data structure of a GCompound to speed up these queries.
        for gobj in self._top:
            if (x, y) in gobj:
                return gobj
        return None

    def _request_focus(self):
        """Ask the OS to assign keyboard focus to this GWindow.

        This brings it to the top and ensures that key events are delivered correctly.
        Clicking in the window automatically requests the focus.

        It is not guaranteed that the OS will give focus to this GWindow.
        """
        _platform.Platform().gwindow_request_focus(self)

    def _repaint(self):
        """Schedule a repaint on this window."""
        _platform.Platform().gwindow_repaint(self)


# NOTE(sredmond): This function exists in this module only so that students don't
# have to import anything special to get a pause function. It should eventually
# be deduplicated into the timer module, and loaded on package import.
def pause(milliseconds):
    """Pause for the given number of milliseconds.

    This is useful for animation where the graphical updates would otherwise be
    too fast to observe.

    To pause for half a second::

        pause(500)

    :param milliseconds: The number of milliseconds to pause.
    """
    _platform.Platform().gtimer_pause(milliseconds)


def screen_width():
    """Return the width of the entire display screen.

    :returns: The width of the display screen.
    """
    return _platform.Platform().gwindow_get_screen_width()


def screen_height():
    """Return the height of the entire display screen.

    :returns: The height of the display screen in pixels.
    """
    return _platform.Platform().gwindow_get_screen_height()


def exit_graphics():
    """Close all graphical windows and forcibly exit the application."""
    # TODO(sredmond): When would we ever want to do this?
    _platform.Platform().gwindow_exit_graphics()



"""A :class:`GBufferedImage` wraps an image and buffers its pixel data.

"""
import campy.datastructures.grid as _grid
import campy.graphics.gcolor as _gcolor
import campy.graphics.gtypes as _gtypes
import campy.private.platform as _platform
import campy.io.base64helper as _base64helper


GBUFFEREDIMAGE_DEFAULT_DIFF_PIXEL_COLOR = 0xdd00dd

# def _char_to_hex(ch):
#      return ord('0') <= ch <= ord('9') ? (ch - ord('0')) : (ch - ord('a') + 10)

# TODO(sredmond): Why on earth is this an interactor?
# class GBufferedImage(_ginteractors.GInteractor):
class GBufferedImage:
    WIDTH_HEIGHT_MAX = 65535

    @classmethod
    def load_from_file(cls, filename):
        inst = cls()
        inst.load(filename)
        return inst

    @property
    def bounds(self):
        return _gtypes.GRectangle(self.x, self.y, self.width, self.height)

    def clear(self):
        self.fill(self.bg_color)

    def count_diff_pixels(self, other, bounds=None):
        # TODO(sredmond): This API differs from the CPP lib.
        # TODO(sredmond): Include bounds
        min_width = min(self.width, other.width)
        min_height = min(self.height, other.height)
        overlap = min_width * min_height
        diff_pixel_count = (self.width * self.height - overlap) + (other.width * other.height - overlap)
        for y in range(min_height):
            for x in range(min_width):
                if self.pixels.get(y, x) != other.pixels.get(y, x):
                    diff_pixel_count += 1
        return diff_pixel_count

    def diff(self, other, diff_color=GBUFFEREDIMAGE_DEFAULT_DIFF_PIXEL_COLOR):
        pass

    def fill(self, color):
        # color = _gcolor.normalize(color)
        self.pixels.fill(color)
        _platform.Platform().gbufferedimage_fill(self, color) # TODO(get rgb value)
        pass

    def fill_region(self, x, y, width, height, TODO):
        pass

    def from_grid(self, grid):
        pass

    def get_height(self):
        pass

    def get_rgb(self, x, y):
        pass

    def get_rgb_string(self, x, y):
        pass

    def get_width(self):
        pass

    def in_bounds(self, x, y):
        pass

    def load(self, filename):
        # TODO(sredmond): Ensure file exists.
        result = _platform.Platform().gbufferedimage_load(self, filename)
        decoded = _base64helper.decode(result)
        self._pixel_string_to_grid(decoded)

    def resize(self, width, height, retain=True):
        pass

    def save(self, filename):
        pass

    def set_rgb(x, y, TODO):
        pass

    def to_grid(self):
        pass

    def _pixel_string_to_grid(self, decoded):
        # Read width (2B) and height (2B)
        # TODO(sredmond): Seriously check this bit fiddling
        w = (decoded[0] << 8) | decoded[1];
        h = (decoded[2] << 8) | decoded[3];
        if w != self.pixels.width() or h != self.pixels.height():
            self.pixels.resize(h, w, retain=False)

        # Check that we received approximately the right length of bytes.
        expected = (w * h * 3) + 4
        actual = len(decoded)
        if actual != decoded:
            print('Expected {} but saw {} bytes'.format(expected, actual))
            # TODO(sredmond): Abort abort
            import sys
            sys.exit(1)

        pixels = []
        for red, green, blue in zip(decoded[4::3], decoded[5::3], decoded[6::3]):
            pixels.append(Pixel(red=red, green=green, blue=blue))

        i = 0
        for row in range(h):
            for col in range(w):
                self.pixels.set(row, col, pixels[i])
                i += 1

        self.width = pixels.width
        self.height = pixels.height

    # "Private" constructor, called by class method constructors.
    def __init__(self, x=0, y=0, width=1, height=1, bg_color=0x000000):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.bg_color = bg_color
        self.pixels = _grid.Grid()

    def __str__(self):
        return 'GBufferedImage()'



"""Interact with the graphics libraries via key events.

NOTE(sredmond): This module is still heavily under development, and should be
considered incomplete by all users.

Usage::

    @campy.onkeypress('a', modifier=SHIFT | CTRL)
    campy.keypress('a')
"""
# See: https://effbot.org/tkinterbook/tkinter-events-and-bindings.htm
import campy.private.platform as _platform

import enum

@enum.unique
class Modifier(enum.Enum):
    SHIFT_DOWN     = 1 << 0
    CTRL_DOWN      = 1 << 1
    META_DOWN      = 1 << 2
    ALT_DOWN       = 1 << 3
    ALT_GRAPH_DOWN = 1 << 4
    BUTTON1_DOWN   = 1 << 5
    BUTTON2_DOWN   = 1 << 6
    BUTTON3_DOWN   = 1 << 7

@enum.unique
class KeyCodes(enum.Enum):
    BACKSPACE_KEY = 8
    TAB_KEY = 9
    ENTER_KEY = 10
    CLEAR_KEY = 12
    ESCAPE_KEY = 27
    PAGE_UP_KEY = 33
    PAGE_DOWN_KEY = 34
    END_KEY = 35
    HOME_KEY = 36
    LEFT_ARROW_KEY = 37
    UP_ARROW_KEY = 38
    RIGHT_ARROW_KEY = 39
    DOWN_ARROW_KEY = 40
    F1_KEY = 112
    F2_KEY = 113
    F3_KEY = 114
    F4_KEY = 115
    F5_KEY = 116
    F6_KEY = 117
    F7_KEY = 118
    F8_KEY = 119
    F9_KEY = 120
    F10_KEY = 121
    F11_KEY = 122
    F12_KEY = 123
    DELETE_KEY = 127
    HELP_KEY = 156


def onkeypress(key, modifiers=None):
    pass


def keypress(key, modifiers=None):
    """Fire a key-pressed event."""
    pass


def setup():
    # self._window.bind("<Key>", lambda event: self.__callback_key_pressed(event))
    pass


# def onkeypress(letter):
#     foobar

# @campy.register(letter, function)



# TODO: rewrite the functinos in this module to be more Pythonic

import enum as _enum

from campy.system.error import error

@_enum.unique
class TokenType(_enum.IntEnum):
    """The enumerated values of the <code>get_token_type</code> method."""
    SEPARATOR = 0
    WORD = 1
    NUMBER = 2
    STRING = 3
    OPERATOR = 4


class TokenScanner():
    def __init__(self, input_stream):
        self.input_stream = input_stream
        self._init_scannner()

    def set_input(self, input_stream):
        self.input_stream = input_stream

    def has_more_tokens(self):
        token = self.next_token()
        self.save_token(token)
        return token != ""

    def next_token(self):
        if self._saved_tokens:
            return self._saved_tokens.pop(0)

        while True:
            if self.ignore_whitespace:
                self._skip_spaces()
            ch = self.input_stream.read(1)

            if ch == '/' and self._ignore_comments:
                ch = self.input_stream.read(1)
                if ch == '/':
                    while True:
                        ch = self.input_stream.read(1)
                        if ch in ('', '\r', '\n'):
                            break
                    continue
                elif ch == '*':
                    prev = ''
                    while True:
                        ch = self.input_stream.read(1)
                        if not ch or (prev == '*' and ch == '/'):
                            break
                        prev = ch
                    continue
                if ch:
                    self.unget_char(ch)
                ch = '/'
            if not ch:
                return ''
            if ch in ('"', "'") and self._scan_strings:
                self.unget_char(ch)
                return self._scan_string()
            if ch.isdigit() and self._scan_numbers:
                self.unget_char(ch)
                return self._scan_number()
            if self.is_word_character(ch):
                self.unget_char(ch)
                return self._scan_word()
            op = ch
            while self._is_operator_prefix(op):
                ch = self.input_stream.read(1)
                if not ch:
                    break
                op += ch
            while len(op) > 1 and not self._is_operator(op):
                self.unget_char(op[-1])
                op = op[:-1]
            return op


    def save_token(self, token):
        self._saved_tokens.append(token)

    def get_position(self):
        pass

    def ignore_whitespace(self):
        pass

    def ignore_comments(self, python_style=True, c_style=False):
        pass

    def scan_numbers(self):
        pass

    def scan_strings(self):
        pass

    def add_word_characters(self, characters):
        self._word_characters += characters

    def is_word_character(self, ch):
        return ch.isalnum() or ch in self._word_characters

    def add_operator(self, op):
        self._operators.append(op)

    def verify_token(self, expected):
        token = self.next_token()
        if token != expected:
            # TODO: error w/ buffer
            error('Found "{}" when expecting "{}"'.format(token, expected))


    def get_token_type(self, token):
        if not token:
            error('Empty token: TODO(EOF)?')

        ch = token[0]
        if ch.isspace():
            return TokenType.SEPARATOR
        elif ch == '"' or (ch == "'" and len(token) > 1):
            return TokenType.STRING
        elif ch.isdigit():
            return TokenType.NUMBER
        elif self.is_word_character(ch):
            return TokenType.WORD
        else:
            return TokenType.OPERATOR

    def get_char(self):
        return self.input_stream.read(1)

    def unget_char(self, ch):
        # TODO: char must match current location
        current = self.input_stream.tell()
        if current > 0:
            self.input_stream.seek(current - 1)

    def get_string_value(self, token):
        out = ''
        start = 0
        finish = len(token)
        if finish > 1 and (token[0] == '"' or token[0] == "'"):
            start = 1
            finish -= 1

        for i in range(start, finish):
            ch = token[i]
            if ch == '\\':
                i += 1
                ch = token[i]
                if ch.isdigit() or ch == 'x':
                    base = 8
                    if ch == 'x':
                        base = 16
                        i += 1
                    result = 0
                    digit = 0
                    while i < finish:
                        ch = token[i]
                        if ch.isdigit():
                            digit = ord(ch) - ord(0)
                        elif ch.isalpha():
                            digit = ord(ch.upper()) - ord('A') + 10
                        else:
                            digit = base
                        if digit >= base:
                            break
                        result = base * result + digit
                        i += 1
                    ch = chr(result)
                    i -= 1
                else:
                    if ch == 'a': ch = '\a'
                    elif ch == 'b': ch = '\b'
                    elif ch == 'f': ch = '\f'
                    elif ch == 'n': ch = '\n'
                    elif ch == 'r': ch = '\r'
                    elif ch == 't': ch = '\t'
                    elif ch == 'v': ch = '\v'
                    # TODO: other delims?
            out += ch
        return out


    # Private
    def _init_scannner(self):
        self._buffer = None
        self._isp = None
        self._string_input = False
        self._ignore_whitespace = False
        self._ignore_comments = False
        self._scan_numbers = False
        self._scan_strings = False
        self._word_characters = []
        self._saved_tokens = []
        self._operators = []

    def _skip_spaces(self):
        while True:
            ch = self.input_stream.read(1)
            if not ch:
                return
            if not ch.isspace():
                self.unget_char(ch)
                return

    def _scan_word(self):
        token = ''
        while True:
            ch = self.input_stream.read(1)
            if not ch:
                break
            if not self.is_word_character(ch):
                self.unget_char(ch)
                break
            token += ch
        return token

    def _scan_number(self):
        token = ''
        state = _NumberScannerState.INITIAL_STATE
        while state != _NumberScannerState.FINAL_STATE:
            ch = self.input_stream.read(1)
            if state == _NumberScannerState.INITIAL_STATE:
                if not ch.isdigit():
                    error('internal error: illegal call')
                state = _NumberScannerState.BEFORE_DECIMAL_POINT
            elif state == _NumberScannerState.BEFORE_DECIMAL_POINT:
                if ch == '.':
                    state = _NumberScannerState.AFTER_DECIMAL_POINT
                elif ch in ('e', 'E'):
                    state = _NumberScannerState.STARTING_EXPONENT
                elif not ch.isdigit():
                    if ch:
                        self.unget_char(ch)
                    state = _NumberScannerState.FINAL_STATE
            elif state == _NumberScannerState.AFTER_DECIMAL_POINT:
                if ch in ('e', 'E'):
                    state = _NumberScannerState.STARTING_EXPONENT
                elif not ch.isdigit():
                    if ch:
                        self.unget_char(ch)
                    state = _NumberScannerState.FINAL_STATE
            elif state == _NumberScannerState.STARTING_EXPONENT:
                if ch in ('-', '+'):
                    state = _NumberScannerState.FOUND_EXPONENT_SIGN
                elif ch.isdigit():
                    state = _NumberScannerState.SCANNING_EXPONENT
                else:
                    if ch:
                        self.input_stream.unget_char(ch)
                    self.input_stream.unget_char(ch)
                    state = _NumberScannerState.FINAL_STATE
            elif state == _NumberScannerState.FOUND_EXPONENT_SIGN:
                if ch.isdigit():
                    state = _NumberScannerState.SCANNING_EXPONENT
                else:
                    if ch:
                        self.input_stream.unget_char(ch)
                    self.input_stream.unget_char(ch)
                    self.input_stream.unget_char(ch)
                    state = _NumberScannerState.FINAL_STATE
            elif case == _NumberScannerState.SCANNING_EXPONENT:
                if not ch.isdigit():
                    if ch:
                        self.input_stream.unget_char(ch)
                    state = _NumberScannerState.FINAL_STATE
            else:
                state = _NumberScannerState.FINAL_STATE
            if state != _NumberScannerState.FINAL_STATE:
                token += ch
        return token

    def _scan_string(self):
        token = ''
        delim = self.input_stream.read(1)
        token += delim
        escape = False
        while True:
            ch = self.input_stream.read(1)
            if not ch:
                error('found unterminated string') # TODO: fn name
            if ch == delim and not escape:
                break
            escape = ch == '\\' and not escape
            token += ch
        return token

    def _is_operator(self, op):
        return op in self._operators

    def _is_operator_prefix(self, op):
        return any(operator.startswith(op) for operator in self._operators)

@_enum.unique
class _NumberScannerState(_enum.IntEnum):
    INITIAL_STATE = 0
    BEFORE_DECIMAL_POINT = 1
    AFTER_DECIMAL_POINT = 2
    STARTING_EXPONENT = 3
    FOUND_EXPONENT_SIGN = 4
    SCANNING_EXPONENT = 5
    FINAL_STATE = 6



"""Top-level Exceptions and Errors.

Provides a generic function to raise an error from a message.

All exceptions (that are not builtin exceptions like ValueError, IndexError)
raised in this library inherit from :class:`CampyException`, the root of all
exceptions in this library.
"""
# TODO(sredmond): Should this be called ErrorException for consistency with ACM?
class CampyException(Exception):
    """Root of all exceptions in the :mod:`campy` module.

    This allows for consistent exception reporting in the module.
    """


class InterruptedIOError(CampyException):
    """A blocking I/O call is interrupted by closing the program."""


def error(message):
    """Generic function to raise an error.

    Signals an error condition in a program by throwing an
    :class:`CampyException`: with the specified message.

    :param str message: error message
    :raises: An :class:`CampyException` with the supplied error message.

    Usage::

        if 'Red Leicester' not in cheeses:
            error("I'm afraid we're fresh out of Red Leicester, sir.")

    You can check whether a code block raises a :class:`CampyException` with::

        try:
            do_something_that_might_raise_a_campy_exception()
        except CampyException:
            recover_from_an_error()

    If the supplied message is an instance or a subclass of :class:`Exception`,
    the :class:`CampyException` is raised from the supplied Exception so that
    traceback information is maintained.
    """
    if isinstance(message, Exception) or issubclass(message, Exception):
        raise CampyException(message) from message
    else:
        raise CampyException(message)


__all__ = ['CampyException', 'InterruptedIOError', 'error']



"""A basic, stripped down queue.

"""
import collections as _collections
import collections.abc as _collections_abc
import functools as _functools


# Implementation note: The head of the queue at index 0.
@_functools.total_ordering
# TODO(sredmond): Sized,Iterable,Container is called Collection in 3.6+
class BasicQueue(_collections_abc.Sized, _collections_abc.Iterable, _collections_abc.Container):
    def __init__(self, data=None):
        self._d = _collections.deque()
        if data:
            self._d.extend(data)

    def __contains__(self, val):
        return val in self._d

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)

    # Methods from C++ library
    def enqueue(self, value):
        self._d.append(value)

    def dequeue(self):
        if not self._d:
            pass
            # error("Queue::dequeue: Attempting to dequeue an empty queue")
        return self._d.popleft()

    def clear(self):
        self._d.clear()

    def peek(self):
        if not self._d:
            pass
            # error("Queue::dequeue: Attempting to dequeue an empty queue")
        return self._d[0]

    def back(self):
        if not self._d:
            pass
            # error("Queue::dequeue: Attempting to dequeue an empty queue")
        return self._d[-1]

    def __str__(self):
        return str(self._d).replace('deque', 'BasicQueue', 1)

    # TODO(sredmond): Add a repr method?

    def __eq__(self, other):
        return self._d == other._d

    def __le__(self, other):
        # TODO(sredmond): There are going to be a lot of problems with unorderable types.
        # TODO(sredmond): Isn't this already the default behavior of __le__?
        return self._d < other._d

    # Synonyms
    add = enqueue
    remove = dequeue
    front = peek
    # Removed: isempty, size, equals

__all__ = ['BasicQueue']



"""Functions for encoding and decoding binary data in the base64 format.

For information on the base64 format, see:

    http://en.wikipedia.org/wiki/Base64

It is a lightweight wrapper around the builtin base64 library that ships with
Python.
"""

import base64 as _base64

def encode(b):
    """Encode bytes-like object b using the standard Base64 alphabet and return the encoded bytes.

    b is the byte string to encode. The encoded byte string is returned.
    """
    return _base64.standard_b64encode(b)

def decode(b):
    """Decode bytes-like object or ASCII string s using the standard Base64 alphabet and return the decoded bytes.

    b is the byte string to decode. The decoded byte string is returned.
    """
    return _base64.standard_b64decode(b)

__all__ = ['encode', 'decode']



"""Load a campy graphical backend.

The default graphical backend is Tk, but others are available in varying
degrees of support.

The available backends are:

- Tk (default): Cross-platform graphics and widget library.
- JBE: Text-based communication to a Java backend: `acm.jar`. Requires Java.

This structure was modelled off of matplotlib's - for more information, see:

https://matplotlib.org/faq/usage_faq.html#what-is-a-backend

There are two ways to configure your backend. If there's a conflict, the last
one in the list will be used.

1. Setting the `CAMPY_BACKEND` environmental variable, either for your current
shell or for a single script::

    $ export CAMPY_BACKEND="Tk"
    $ python test.py

    $ CAMPY_BACKEND="tk" python test.py

2. If your script depends on a specific backend you can call `use()`::

    $ import campy
    $ campy.use('tk')

The `use()` function must be invoked before importing any other campy modules.
Calling `use` after another `campy` module has been imported will have no
effect. If you are building library code, you should avoid explicitly calling
`use()` unless absolutely necessary because your users will have to change the
code if they want to use a different backend.

Note: Backend name specifications are not case-sensitive, e.g. 'Tk' and 'tk'
are equivalent.
"""
# TODO(sredmond): Add support for `use`.
# TODO(sredmond): Add support for a .campyrc file.
from campy.private.backends.jbe.backend_jbe import JavaBackend
from campy.private.backends.tk.backend_tk import TkBackend

import logging
import os

# Module-level logger.
logger = logging.getLogger(__name__)

# The default backend to use when an environmental variable doesn't override.
DEFAULT_BACKEND = 'Tk'

backend_name = os.environ.get('CAMPY_BACKEND', DEFAULT_BACKEND).lower()
logger.debug('Attempting to create backend {!r}'.format(backend_name))

# Load a backend.
# TODO(sredmond): This is effectively a singleton because Python only imports
# modules once, unless forced to otherwise, such as with Jupyter's %autoreload
# magic. To defend against forced module reimport, this should be encapsulated.
if backend_name == 'tk':  # Load the Tk backend
    backend = TkBackend()
elif backend_name == 'jbe':  # Load the JBE backend.
    backend = JavaBackend()
else:
    raise ImportError('Unrecognized backend: {!r}'.format(backend_name))



#!/usr/bin/env python3 -tt
"""
File: simpio.h
--------------
This file exports a set of functions that simplify input/output
operations in Python and provide some error-checking on console input.

Modified from Marty Stepp's CPP libraries.

@author sredmond
"""

GETINTEGER_DEFAULT_PROMPT = "Enter an integer: ";
GETINTEGER_DEFAULT_REPROMPT = "Illegal integer format. Try again.";
GETREAL_DEFAULT_PROMPT = "Enter a number: ";
GETREAL_DEFAULT_REPROMPT = "Illegal numeric format. Try again.";
GETPOSITIVEREAL_DEFAULT_PROMPT = "Enter a positive number: "
GETPOSITIVEREAL_DEFAULT_REPROMPT = "Illegal numeric format. Try again. "
GETYESORNO_DEFAULT_PROMPT = "Yes or No? ";
GETYESORNO_DEFAULT_REPROMPT = "Please type a word that starts with 'Y' or 'N': ";
DEFAULT_PROMPT = "> "
DEFAULT_REPROMPT = "Invalid input. Please try again. "


def get_integer(prompt=GETINTEGER_DEFAULT_PROMPT, reprompt=GETINTEGER_DEFAULT_REPROMPT):
    return get_fn_cond(
        lambda line: int(line.strip()),
        lambda _: True,
        prompt, reprompt
    )


def get_real(prompt=GETREAL_DEFAULT_PROMPT, reprompt=GETREAL_DEFAULT_REPROMPT):
    return get_fn_cond(
        lambda line: float(line.strip()),
        lambda _: True,
        prompt, reprompt
    )


def get_positive_real(prompt=GETPOSITIVEREAL_DEFAULT_PROMPT, reprompt=GETPOSITIVEREAL_DEFAULT_REPROMPT):
    return get_fn_cond(
        lambda line: float(line.strip()),
        lambda val: val > 0,
        prompt, reprompt
    )


def get_yes_or_no(prompt=GETYESORNO_DEFAULT_PROMPT, reprompt=GETYESORNO_DEFAULT_REPROMPT, default=None):
    result = get_fn_cond(
        lambda line: line[0].upper() if line else default,
        lambda val: val in ['Y', 'N'],
        prompt, reprompt
    )
    return result == 'Y'


def get_fn_cond(fn, pred, prompt=DEFAULT_PROMPT, reprompt=DEFAULT_REPROMPT):
    prompt = append_space(prompt)
    while True:
        line = input(prompt)
        try:
            out = fn(line)
        except Exception:
            pass
        else:
            if pred(out):
                return out
        if reprompt:
            print(append_space(reprompt))


def get_line(prompt=None):
    return input(append_space(prompt))


def append_space(prompt):
    """Adds a space to the end of the given string if none is present."""
    if not prompt.endswith(' '):
        return prompt + ' '
    return prompt



"""Interact with the graphics libraries via mouse events.

- MOUSE_CLICKED
- MOUSE_PRESSED
- MOUSE_RELEASED
- MOUSE_MOVED
- MOUSE_DRAGGED
"""
import campy.private.platform as _platform

import enum

class GMouseEvent:
    """An event generated by some mouse interaction.

    Each mouse event records its event type, its parent window, and the mouse
    coordinate of the event.



    # Each mouse event
    # This event subclass represents a mouse event.  Each mouse event
    # records the event type (MOUSE_PRESSED,
    # MOUSE_RELEASED, MOUSE_CLICKED,
    # MOUSE_MOVED, MOUSE_DRAGGED) along
    # with the coordinates of the event.  Clicking the mouse generates
    # three events in the following order: MOUSE_PRESSED,
    # MOUSE_RELEASED, MOUSE_CLICKED.

    # As an example, the following program uses mouse events to let
    # the user draw rectangles on the graphics window.  The only
    # complexity in this code is the use of the library functions
    # min and abs to ensure that the
    # dimensions of the rectangle are positive::

    #     gw = _gwindow.GWindow
    #     print("This program lets the user draw rectangles.")
    #     while(True):
    #         e = gevents.waitForEvent();
    #         if (e.getEventType() == MOUSE_PRESSED):
    #             startX = e.getX();
    #             startY = e.getY();
    #             rect = gobjects.GRect(startX, startY, 0, 0);
    #             rect.setFilled(True);
    #             gw.add(rect);
    #         elif(e.getEventType() == MOUSE_DRAGGED):
    #             x = min(e.getX(), startX);
    #             y = min(e.getY(), startY);
    #             width = abs(e.getX() - startX);
    #             height = abs(e.getY() - startY);
    #             rect.setBounds(x, y, width, height);

    # Attributes:
    #     gwindow [GWindow]: GWindow in which MouseEvent occurred.
    #     x [float]: The x-coordinate at which the event occurred.
    #     y [float]: The y-coordinate at which the event occurred.

    # Notes:
    #     The x- and y-coordinates are given relative to the window origin at
    #     the upper left corner of the window.
    """

    def __init__(self, event_type, gwindow, x, y):
        """Create a :class:`GMouseEvent` with the given arguments.

        :param event_type: Which type of mouse event this object represents.
        :param gwindow: The originating :class:`GWindow`.
        :param x: The x-coordinate of the mouse.
        :param y: The y-coordinate of the mouse.
        """
        super().__init__()
        self._type = event_type
        self._gwindow = gwindow
        self._x = x
        self._y = y

    @property
    def window(self):
        return self._gwindow

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    def __str__(self):
        return "GMouseEvent({}, x={}, y={})".format(self.event_type.name, self.x, self.y)


@enum.unique
class MouseEventType(enum.Enum):
    MOUSE_CLICKED = 1  # TODO(sredmond): This is really a MOUSE_PRESSED. Note that somewhere.
    MOUSE_RELEASED = 2
    MOUSE_MOVED = 3
    MOUSE_DRAGGED = 4
    # MOUSE_ENTERED = 5
    # MOUSE_EXITED = 6
    # MOUSE_DOUBLE_CLICKED = 7
    # MOUSE_TRIPLE_CLICKED = 8


def onmouseclicked(function):
    """
    Usage::

        %gui tk
        from campy.gui.events.mouse import *
        from campy.graphics.gobjects import *
        from campy.graphics.gwindow import *

        window = GWindow()

        @onmouseclicked
        def add_circle(event):
            event.window.add(GOval(100, 100, x=event.x - 50, y=event.y - 50))
    """
    _platform.Platform().event_add_mouse_handler(MouseEventType.MOUSE_CLICKED, function)


def onmousereleased(function):
    """
    Usage::

        %gui tk
        from campy.gui.events.mouse import *
        from campy.graphics.gobjects import *
        from campy.graphics.gwindow import *

        window = GWindow()

        @onmousereleased
        def add_rect(event):
            event.window.add(GRect(100, 100, x=event.x - 50, y=event.y - 50))
    """
    _platform.Platform().event_add_mouse_handler(MouseEventType.MOUSE_RELEASED, function)


def onmousemoved(function):
    """
    Usage::

        %gui tk
        from campy.gui.events.mouse import *
        from campy.graphics.gobjects import *
        from campy.graphics.gwindow import *

        window = GWindow()

        @onmousemoved
        def add_circle(event):
            event.window.add(GOval(100, 100, x=event.x - 50, y=event.y - 50))
    """
    _platform.Platform().event_add_mouse_handler(MouseEventType.MOUSE_MOVED, function)


def onmousedragged(function):
    """
    Usage::

        %gui tk
        from campy.gui.events.mouse import *
        from campy.graphics.gobjects import *
        from campy.graphics.gwindow import *

        window = GWindow()

        @onmousedragged
        def add_circle(event):
            event.window.add(GOval(100, 100, x=event.x - 50, y=event.y - 50))
    """
    _platform.Platform().event_add_mouse_handler(MouseEventType.MOUSE_DRAGGED, function)


# TODO(sredmond): Add code to generate mouse events (for mocking) too.

# Convenient testing code.
# from campy.graphics.gwindow import GWindow

# from campy.graphics.gobjects import GOval, GRect

# from campy.gui.events.mouse import *

# window = GWindow(title='Mouse Event Test')

# @onmouseclicked
# def add_rect_on_click(event):
#         event.window.add(GRect(100, 100, x=event.x - 50, y=event.y - 50))

# @onmousereleased
# def add_circle_on_release(event):
#         event.window.add(GOval(100, 100, x=event.x - 50, y=event.y - 50))

# @onmousemoved
# def add_circle_on_move(event):
#         event.window.add(GOval(50, 50, x=event.x - 25, y=event.y - 25))

# @onmousedragged
# def add_rect_on_drag(event):
#         event.window.add(GRect(50, 50, x=event.x - 25, y=event.y - 25))




"""
CAMPY Graphics Library
~~~~~~~~~~~~~~~~~~~~~~

The campy library makes Stanford's graphical ACM libraries available in Python.

# TODO(sredmond): Add basic usage examples.

Full documentation is available at <https://campy.sredmond.io/>

:copyright: (c) 2016-2019 by Sam Redmond.
:license: MIT License. See LICENSE for more details.
"""
import warnings

# TODO(sredmond): Check compatibility of dependency versions.

# TODO(sredmond): Check backend availability.
try:
    import _tkinter
except ImportError as err:
    warnings.warn('Unable to find _tkinter library. If using the Tk backend, you will need to install a version of Python with Tk support.')

try:
    import tkinter
except ImportError as err:
    warnings.warn('Unable to find tkinter library. If using the Tk backend, you will need to install a version of Python with Tk support.')

# Import package metadata.
from .__version__ import (
    __title__, __description__, __url__, __license__,
    __version__, __build__, __status__, __author__,
    __maintainer__, __email__, __copyright__, __credits__,
    __snake__
)

# OPTIONAL(sredmond): Import subpackages to the top-level namespace.

# Set default logging handler to avoid "No handler found" warnings.
import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())  # Add a default do-nothing handler to these logs.

# TODO(sredmond): FOR DEVELOPMENT ONLY. Log all messages.
import os
if os.environ.get('CAMPY_DEVELOP'):
    default_log_level = 'DEBUG'
else:
    default_log_level = 'WARNING'
level = os.environ.get('CAMPY_LOGLEVEL', default_log_level).upper()

logging.basicConfig()
logger.setLevel(level)
# Send some initial messages.
logger.info("Welcome to the campy libraries.")
logger.info("If you have any questions, reach out to {me} at {email}".format(me=__maintainer__, email=__email__))



"""
File: sound.py
--------------
This file defines a class that represents a sound.
"""
import campy.private.platform as _platform

class Sound():
    """
    This class encapsulates a sound file.  The sound file is specified in the
    constructor and must be a file in either the current directory or a
    subdirectory named sounds.

    The following code, for example, plays the sound file ringtone.wav::

        ringtone = Sound("ringtone.wav")
        ringtone.play()

    """
    def __init__(self, filename):
        """Creates a Sound object.
        The default constructor creates an empty sound that cannot be played.
        The second form initializes the sound by reading in the contents of
        the specified file.

        @type filename: string
        @param filename: sound file
        @rtype: void
        """
        self.filename = filename
        _platform.Platform().create_sound(self, filename)


    def play(self):
        """Starts playing the sound.

        This call returns immediately without waiting for the sound to finish.
        """
        _platform.Platform().play_sound(self)

    def __del__(self):
        """Frees the Java memory associated with the sound."""
        _platform.Platform().delete_sound(self)



"""
This file exports a standardized set of tools for working with
files. The library offers at least some portability across the
file systems used in the three supported platforms: Mac OSX,
Windows, and Linux.  Directory and search paths are allowed to
contain separators in any of the supported styles, which usually
makes it possible to use the same code on different platforms.

Note: Several of the functions that lived in CPP's version of this
library are removed, to force students to learn the proper builtin
pathlib library. The compatability is listed here::

    def create_directory(path):
        _Path(path).mkdir()

    def create_directory_path(path):
        _Path(path).mkdir(path, parents=True)

    def default_extension(path, ext):
        p = _Path(path)
        if ext.startswith('*') or not p.suffix:
            return p.with_suffix(ext)

    def delete_file(path):
        return os.remove(path)

TODO more
"""
from pathlib import Path as _Path, PurePath as _PurePath
import campy.private.platform as _platform

def expand_pathname(path):
    pass

def file_exists(path):
    pass

def find_on_path(path, filename):
    pass

def get_current_directory():
    pass

def get_directory_path_separator():
    pass

def get_extension(path):
    pass

def get_head(path):
    pass

def get_root(path):
    pass

def get_search_path_separator(path):
    pass

def get_tail(path):
    pass

def get_temp_directory():
    pass

def is_directory(path):
    pass

def is_file(path):
    pass

def is_symbolic_link(path):
    pass

def list_directory(path):
    pass

def match_filename_pattern(path):
    pass

def open_file(path):
    pass

def open_file_dialog(title):
    pass

def open_on_path(path):
    pass

def prompt_user_for_file(prompt="", reprompt=""):
    pass

def read_entire_file(stream, lines):
    pass

def rename(old, new):
    pass

def rewindStream(input):
    pass

def set_current_directory(path):
    pass

def write_entire_file(path):
    pass


# '''
# This file exports a standardized set of tools for working with
# files.  The library offers at least some portability across the
# file systems used in the three supported platforms: Mac OSX,
# Windows, and Linux.  Directory and search paths are allowed to
# contain separators in any of the supported styles, which usually
# makes it possible to use the same code on different platforms.
# '''
# import platform
# import os
# import exceptions

# if(os.name == "posix"): import pwd

# def expandPathname(filename):
#   '''
#   Expands a filename into a canonical name for the platform.

#   @type filename: string
#   @param filename: filename to expand
#   @rtype: string
#   '''
#   if(filename == ""): return ""
#   length = len(filename)
#   if(os.name == "posix"):
#       strPos = 1
#       while(strPos < length \
#           and filename[strPos] != "\\" \
#           and filename[strPos] != "/"):
#           strPos += 1

#       homedir = None
#       if(strPos == 1):
#           homedir = os.getenv("HOME")
#           if(homedir == None): homedir = pwd.getpwuid(os.getuid()).pw_dir
#       else:
#           pw = pwd.getpwnam(filename[1:strPos])
#           if(pw == None):
#               raise exception.StandardError
#           homedir = pw.pw_dir

#       filename = homedir + filename[strPos:]
#       length = len(filename)

#       for i in range(length):
#           if(filename[i] == "\\"):
#               filename = filename[:i] + "/" + filename[i+1:]

#   else:
#       for i in range(length):
#           if(filename[i] == "/"):
#               filename = filename[:i] + "\\" + filename[i+1:]

#   return filename

# def openFile(filename, binary=False):
#   '''
#   Opens the filestream stream using the specified
#   filename.  This function is similar to the open
#   method of the stream classes, but uses a C++ string
#   object instead of the older C-style string.  If the operation
#   succeeds, openFile returns true;
#   if it fails, openFile sets the failure flag in the
#   stream and returns false.

#   @type filename: string
#   @param filename: file to open
#   @type binary: boolean
#   @param binary: open file in binary mode on Windows
#   @rtype: file
#   '''
#   mode = "r+"
#   if(binary):
#       mode = "r+b"

#   file = None

#   try:
#       file = open(expandPathname(filename), mode)
#   except IOError:
#       file = None

#   return file

# def promptUserForFile(prompt, binary=False):
#   '''
#   Asks the user for the name of a file.  The file is opened using
#   the reference parameter stream, and the function
#   returns the name of the file.  If the requested file cannot be
#   opened, the user is given additional chances to enter a valid file.
#   The optional prompt argument provides an input prompt
#   for the user.

#   @type prompt: string
#   @param prompt: file prompt text
#   @type binary: boolean
#   @param binary: open file in binary mode on Windows
#   @rtype: file
#   '''
#   while(True):
#       filename = raw_input(prompt)
#       file = openFile(filename, binary)
#       if(file != None): return file
#       print("Unable to open that file. Try again.")
#       if(prompt == ""): prompt = "Filename: "

def open_file_dialog(title = "Open File", mode="load", path = "", binary=False):
  '''
  Opens a dialog that allows the user to choose the file.  The
  title parameter is displayed in the dialog title.
  The path parameter is used to set the working directory;
  if path does not appear, openFileDialog
  uses the current directory.

  @type title: string
  @param title: title of dialog box
  @type mode: string
  @param mode: mode of dialog box
  @type binary: boolean
  @param binary: open file in binary mode on windows
  @rtype: file
  '''
  filename = _platform.Platform().openFileDialog(title, mode, path)
  if(filename == ""): return None
  return filename

# def readEntireFile(file):
#   '''
#   Reads the entire contents of the specified input stream into the
#   string vector lines.  The client is responsible for
#   opening and closing the stream.  The vector can be either an STL
#   vector or a Vector as defined in the
#   Stanford C++ libraries.

#   @type file: file
#   @param file: file to read
#   @rtype: [string]
#   @return: list of strings representing all lines in the file
#   '''
#   return file.readlines()

# def getRoot(filename):
#   '''
#   Returns the root of filename.  The root consists
#   of everything in filename up to the last dot and
#   the subsequent extension.  If no dot appears in the final component
#   of the filename, getRoot returns the entire name.

#   @type filename: string
#   @param filename: filename to get root from
#   @rtype: string
#   '''
#   dot = -1
#   length = len(filename)
#   for i in range(length):
#       if(filename[i] == "."): dot = i
#       if(filename[i] == "/" or filename[i] == "\\"): dot = -1

#   if(dot == -1):
#       return filename
#   else:
#       return filename[:dot]

# def getExtension(filename):
#   '''
#   Returns the extension of filename.  The extension
#   consists of the separating dot and all subsequent characters.
#   If no dot exists in the final component, getExtension
#   returns the empty string.  These semantics ensure that concatenating
#   the root and the extension always returns the original filename.

#   @type filename: string
#   @param filename: filename to get extension from
#   @rtype: string
#   '''
#   dot = -1
#   length = len(filename)
#   for i in range(length):
#       if(filename[i] == "."): dot = i
#       if(filename[i] == "/" or filename[i] == "\\"): dot = -1

#   if(dot == -1):
#       return ""
#   else:
#       return filename[dot:]

# def getHead(filename):
#   '''
#   Returns all but the last component of a path name.  The components
#   of the path name can be separated by any of the directory path
#   separators (forward or reverse slashes).  The special cases are
#   illustrated by the following examples:

#       - getHead("a/b")  = "a"     getTail("a/b")   = "b"
#       - getHead("a")    = ""      getTail("a")     = "a"
#       - getHead("/a")   = "/"     getTail("/a")    = "a"
#       - getHead("/")    = "/"     getTail("/")     = ""

#   @type filename: string
#   @param filename: filename to get head from
#   @rtype: string
#   '''
#   slash = -1
#   length = len(filename)
#   for i in range(length):
#       if(filename[i] == "/" or filename[i] == "\\"):
#           slash = i

#   if(slash == -1):
#       return ""
#   elif(slash == 0):
#       return "/"
#   else:
#       return filename[0:slash]

# def getTail(filename):
#   '''
#   Returns the last component of a path name.  The components of the
#   path name can be separated by any of the directory path separators
#   (forward or reverse slashes).  For details on the interpretation of
#   special cases, see the comments for the getHead function.

#   @type filename: string
#   @param filename: filename to get tail from
#   @rtype: string
#   '''
#   slash = -1
#   length = len(filename)
#   for i in range(length):
#       if(filename[i] == "/" or filename[i] == "\\"):
#           slash = i

#   if(slash == -1):
#       return filename
#   else:
#       return filename[slash+1:]

# def defaultExtension(filename, ext):
#   '''
#   Adds an extension to a file name if none already exists.  If the
#   extension argument begins with a leading *,
#   any existing extension in filename is replaced by
#   ext.

#   @type filename: string
#   @param filename: filename to change extension on
#   @type ext: string
#   @param ext: new extension
#   @rtype: string
#   '''
#   force = (ext[0] == "*")
#   if(force): ext = ext[1:]
#   dot = -1
#   length = len(filename)
#   for i in range(length):
#       if(filename[i] == "."): dot = i
#       if(filename[i] == "/" or filename[i] == "\\"): dot = -1

#   if(dot == -1):
#       force = True
#       dot = length

#   if(force):
#       return filename[0:dot] + ext
#   else:
#       return filename

# def openOnPath(path, filename, binary = False):
#   '''
#   Opens a file using a search path.  If openOnPath
#   is successful, it returns the first path name on the search path
#   for which stream.open succeeds.  The path
#   argument consists of a list of directories that are prepended to the
#   filename, unless filename begins with an absolute
#   directory marker, such as / or ~.
#   The directories in the search path may be separated either
#   by colons (Unix or Mac OS) or semicolons (Windows).  If the file
#   cannot be opened, the failure bit is set in the stream
#   parameter, and the openOnPath function returns the
#   empty string.

#   @type path: string
#   @param path: search paths to use when finding filename
#   @type filename: string
#   @param filename: file to search for
#   @type binary: boolean
#   @param binary: open in binary mode on windows
#   @rtype: file
#   '''
#   paths = splitPath(path)
#   for dir in paths:
#       pathname = dir + "/" + filename
#       file = openFile(pathname, binary)
#       if(file != None): return file

#   return None

# def findOnPath(path, filename):
#   '''
#   Returns the canonical name of a file found using a search path.
#   The findOnPath function is similar to
#   openOnPath, except that it doesn't actually
#   return an open stream.  If no matching file is found,
#   findOnPath returns the empty string.

#   @type path: string
#   @param path: search paths to use when finding filename
#   @type filename: string
#   @param filename: file to search for
#   @rtype: string
#   '''
#   file = openOnPath(path, filename)
#   if(file == None):
#       return None

#   file.close()
#   return file.name

# def deleteFile(filename):
#   '''
#   Deletes the specified file.  Errors are reported by calling
#   error.

#   @type filename: string
#   @param filename: file to delete
#   @rtype: void
#   '''
#   os.remove(expandPathname(filename))

# def renameFile(oldname, newname):
#   '''
#   Renames a file.  Errors are reported by calling
#   error in the implementation.

#   @type oldname: string
#   @param oldname: file to rename
#   @type newname: string
#   @param newname: new name for file
#   @rtype: void
#   '''
#   oldname = expandPathname(oldname)
#   newname = expandPathname(newname)
#   os.rename(oldname, newname)

# def createDirectory(path):
#   '''
#   Creates a new directory for the specified path.  The
#   createDirectory function does not report an error if
#   the directory already exists.  Unlike createDirectoryPath,
#   createDirectory does not create missing directories
#   along the path.  If some component of path does
#   not exist, this function signals an error.

#   @type path: string
#   @param path: path to create new directory on
#   @rtype: void
#   '''
#   os.mkdir(path)

# def createDirectoryPath(path):
#   '''
#   Creates a new directory for the specified path.   If intermediate
#   components of path do not exist, this function creates
#   them as needed.

#   @type path: string
#   @param path: path to created new directory on
#   @rtype: void
#   '''
#   cp = 1
#   if(path == ""): return
#   while(true):
#       cp = path.find(os.pardir, cp+1)
#       if(cp == -1): break
#       os.makedirs(path[:cp - 1])
#   os.makedirs(path)

# def fileExists(filename):
#   '''
#   Returns true if the specified file exists.

#   @type filename: string
#   @param filename: file to check existence of
#   @rtype: boolean
#   '''
#   return os.path.exists(filename)

# def isFile(filename):
#   '''
#   Returns true if the specified file is a regular file.

#   @type filename: string
#   @param filename: file to check status of
#   @rtype: boolean
#   '''
#   return os.path.isfile(filename)

# def isSymbolicLink(filename):
#   '''
#   Returns true if the specified file is a symbolic link.

#   @type filename: string
#   @param filename: file to check status of
#   @rtype: boolean
#   '''
#   return os.path.islink(filename)

# def isDirectory(filename):
#   '''
#   Returns true if the specified file is a directory.

#   @type filename: string
#   @param filename: file to check status of
#   @rtype: boolean
#   '''
#   return os.path.isdir(filename)

# def setCurrentDirectory(path):
#   '''
#   Changes the current directory to the specified path.

#   @type path: string
#   @param path: directory to change current working directory to
#   @rtype: boolean
#   @return: directory change succeeded
#   '''
#   return os.chdir(path)

# def getCurrentDirectory():
#   '''
#   Returns an absolute filename for the current directory.

#   @rtype: string
#   @return: path of current working directory
#   '''
#   return os.getcwd()

# def listDirectory(path):
#   '''
#   Adds an alphabetized list of the files in the specified directory
#   to the string vector list.  This list excludes the
#   names . and .. entries.

#   @type path: string
#   @param path: directory to list files in
#   @rtype: [string]
#   @return: alphabetized list of files in directory
#   '''
#   return os.listdir(path).sort()

# def getDirectoryPathSeparator():
#   '''
#   Returns the standard directory path separator used on this platform.

#   @rtype: string
#   '''
#   return os.sep

# def getSearchPathSeparator():
#   '''
#   Returns the standard search path separator used on this platform.

#   @rtype: string
#   '''
#   return os.pathsep

# def matchFilenamePattern(filename, pattern):
#   '''
#   Determines whether the filename matches the specified pattern.  The
#   pattern string is interpreted in much the same way that a Unix shell
#   expands filenames and supports the following wildcard options:

#       - ? Matches any single character
#       - * Matches any sequence of characters
#       - [...]  Matches any of the specified characters
#       - [^...] Matches any character except the specified ones

#   The last two options allow a range of characters to be specified in the
#   form a-z.

#   @type filename: string
#   @param filename: filename to check against pattern
#   @type pattern: string
#   @param pattern: pattern to verify
#   @rtype: boolean
#   '''
#   return recursiveMatch(filename, 0, pattern, 0)

# def splitPath(path):
#   '''
#   private method
#   '''
#   list = []
#   sep = ":" if (path.find(";") == -1) else ";"
#   path += sep
#   start = 0
#   while(True):
#       finish = path.find(sep, start)
#       if(finish == -1): break
#       if(finish > start + 1):
#           list.append(path[start:finish])
#       start = finish + 1

#   return list

# def recursiveMatch(str, sx, pattern, px):
#   '''
#   private method
#   '''
#   slen = len(str)
#   plen = len(pattern)
#   if(px == plen): return (sx == slen)
#   pch = pattern[px]

#   if(pch == "*"):
#       for i in range(sx, slen + 1):
#           if(recursiveMatch(str, i, pattern, px + 1)): return True
#       return False

#   if(sx == slen): return False

#   sch = str[sx]
#   if(pch == "["):
#       match = False
#       invert = False
#       px += 1
#       if(px == plen):
#           raise exception.StandardError # Throw error: missing ]
#           dummy = 1
#       if(pattern[px] == "^"):
#           px += 1
#           invert = True
#       while(px < plen and pattern[px] != "]"):
#           if(px + 2 < plen and pattern[px + 1] == "-"):
#               match = (match or (sch >= pattern[px] and sch <= pattern[px+2]))
#               px += 3
#           else:
#               match = (match or (sch == pattern[px]))
#               px += 1
#       if(px == plen):
#           raise exception.StandardError # throw error: missing ]
#           dummy = 1
#       if(match == invert): return False
#   elif(pch != "?"):
#       if(pch != sch): return False
#   return recursiveMatch(str, sx + 1, pattern, px +1)



# if __name__ == '__main__':
#     pass




"""Classes for manipulating colors and pixels.

A :class:`GColor` is a heavy object representing an abstract color. It
should be thought of as an opaque data type that can be initialized from any of
several possible representations, and serialized to any of those representations.

A :class:`Pixel` represents a single pixel of an image, and is stored much more
efficiently. A Pixel can represent an (R, G, B) triplet or an (R, G, B, A)
quadruplet.

The main difference between a :class:`Pixel` and a :class:`GColor` is that a
:class:`Pixel` is basically an efficiently stored integer representing the alpha,
red, green, and blue channels of the pixel, whereas a :class:`GColor` has no
alpha channel and has many more utility methods attached.

Additionally, a :class:`Pixel` is mutable, whereas a :class:`GColor` is immutable.
"""
# TODO(sredmond): Rethink Pixel mutability, since changes won't propagate to
# Pixel containers.
from campy.system.error import CampyException

import math


class Pixel:
    """A :class:`Pixel` represents a single, possibly-transparent pixel from an image.

    Students will likely never have to directly instantiate Pixels, and will instead
    likely just modify provided Pixels.

    Pixels can be modified directly by getting or setting their properties.

    To change the values of channels of a pixel::

        cardinal = Pixel(168, 0, 59)
        print(cardinal.red)  # => 168

        cardinal.red = 223
        cardinal.green = 202
        cardinal.blue = 151

        red, green, blue = cardinal.rgb()

    To modify the alpha channel of a pixel, use the same syntax::

        opaque = Pixel(168, 0, 59)
        opaque.alpha //= 2
        alpha, red, green, blue = opaque.argb()

    When setting values for these channels, it is required to supply an integer between
    0 and 255, inclusive.
    """
    # TODO(sredmond): Document which methods produce pixels.
    # TODO(sredmond): Revisit subclassing from int.

    # Tell Python to disallow dynamic attribute creation except for self._value
    __slots__ = ['_value']

    def __init__(self, red=0x00, green=0x00, blue=0x00, alpha=0xFF):
        """Create a new Pixel by channels.

        The red, green, blue, and alpha channels can be independently set.

        To create a default (black) pixel::

            default = Pixel()

        To create a pixel with given red, green, and blue values::

            cardinal = Pixel(168, 0, 59)

        To create a red pixel::

            red = Pixel(red=0xFF)

        To create a cyan (half-blue, half-green) pixel::

            cyan = Pixel(green=0x7F, blue=0x7F)

        To create a semi-transparent green pixel::

            pixel = Pixel(green=0xFF, alpha=0x7F)

        :param red: The value of the red channel between 0 and 255.
        :param green: The value of the green channel between 0 and 255.
        :param blue: The value of the blue channel between 0 and 255.
        :param alpha: The value of the alpha channel between 0 and 255.
        """

        # NOTE(sredmond): Since int is immutable, we can't set ourself (or can we?!)
        # TODO(sredmond): Check that the supplied arguments are within valid bounds.
        self._value = (alpha << 24) | (red << 16) | (green << 8) | blue

    @property
    def alpha(self):
        """Get or set this :class:`Pixel`'s alpha channel.

        Usage::

            pixel = Pixel(168, 0, 59)
            pixel.alpha = 127
            print(pixel.alpha)
        """
        # TODO(sredmond): The C++ libraries claim that you "have to & a second time
        # because of sign-extension on >> shift." That's because if the value is
        # stored in something like an int_32, some values of alpha might mean that
        # the integer is negative, so right-shifting would carry over that negativity.
        # This shift doesn't apply in Python, since ints are unbounded.
        return (self._value >> 24) & 0xFF

    @property
    def red(self):
        """Get or set this :class:`Pixel`'s red channel.

        Usage::

            pixel = Pixel(168, 0, 59)
            pixel.red = 41
            print(pixel.red)
        """
        return (self._value >> 16) & 0xFF

    @property
    def green(self):
        """Get or set this :class:`Pixel`'s green channel.

        Usage::

            pixel = Pixel(168, 0, 59)
            pixel.green = 41
            print(pixel.green)
        """
        return (self._value >> 8) & 0xFF

    @property
    def blue(self):
        """Get or set this :class:`Pixel`'s blue channel.

        Usage::

            pixel = Pixel(168, 0, 59)
            pixel.blue = 41
            print(pixel.blue)
        """
        return self._value & 0xFF

    @alpha.setter
    def alpha(self, new_alpha):
        self._value &= ~(0xFF << 24)
        self._value |= new_alpha << 24

    @red.setter
    def red(self, new_red):
        self._value &= ~(0xFF << 16)
        self._value |= (new_red << 16)

    @green.setter
    def green(self, new_green):
        self._value &= ~(0xFF << 8)
        self._value |= new_green << 8

    @blue.setter
    def blue(self, new_blue):
        self._value &= ~0xFF
        self._value |= new_blue

    def rgb(self):
        """Return a 3-tuple of this Pixel's channels, not including alpha transparency.

        The elements of the tuple correspond to red, green, and blue.

        To extract all color channels at once::

            pixel = Pixel(168, 0, 59)
            red, green, blue = pixel.rgb()
        """
        return self.red, self.green, self.blue

    def argb(self):
        """Return a 4-tuple of this Pixel's channels, including alpha transparency.

        The elements of the tuple correspond to alpha, red, green, and blue.

        To extract all channels at once::

            pixel = Pixel(168, 0, 59)
            alpha, red, green, blue = pixel.argb()
        """
        return self.alpha, self.red, self.green, self.blue

    def __str__(self):
        return "Pixel(red={self.red}, green={self.green}, blue={self.blue})".format(self=self)


class _ColorResolverMeta(type):
    """Metaclass to override attribute access for :class:`GColor`.

    """
    def __getattr__(cls, attr):
        """Implement ``self.attr`` if normal attribute lookup fails.

        As a metaclass, this will modify attribute lookup on its derived class.
        """
        # TODO(sredmond): Optionally also remove non-ASCII characters.
        # TODO(Sredmond): "Did you mean... GColor.BLUE" with config flag
        attr = attr.strip().lower().replace(' ', '').replace('_', '').replace('-', '')
        if attr in COLORS:
            # TODO(sredmond): This constructs a new GColor each time.
            return cls.normalize(COLORS[attr])
        raise AttributeError("type object '{}' has no attribute '{}'".format(cls.__name__, attr))


class GColor(metaclass=_ColorResolverMeta):
    """A :class:`GColor` represents a color.

    Colors can be represented in many different forms.

    (1) As a three-integer tuple, such as as (168, 0, 59), with each entry in the range 0..255
    (2) As a three-character hex string prefaced with '#', such as "#83E"
    (3) As a six-character hex string prefaced with '#', such as "#A8003B"
    (4) As a three-character hex string prefaced with '0x', such as "0x83E"
    (5) As a six-character hex string prefaced with '0x', such as "0xA8003B"
    (6) As a case-insensitive name, such as "red" or "LiGhT sKy BlUe"
    (7) As a GColor constant, such as GColor.RED or GColor.LIGHT_SKY_BLUE.
    (8) As a 24-bit integer, representing the red, green, and blue channels.

    The canonical (internal) form for a GColor is as three integers between 0 and 255.

    A :class:`GColor` should be thought of as immutable.
    """

    def __init__(self, red, green, blue):
        self._red = red
        self._green = green
        self._blue = blue

    @classmethod
    def normalize(cls, color):
        """Normalize a color description provided by an end user."""
        if isinstance(color, GColor):
            return color

        if isinstance(color, int):
            # Mode (8): 24-bit integer.
            red = (color >> 16) & 0xFF
            green = (color >> 8) & 0xFF
            blue = color & 0xFF
            return cls(red, green, blue)

        elif isinstance(color, (tuple, list)):
            # Mode (1): 3-element tuple.
            if len(color) != 3:
                raise SomeException()  # TODO(sredmond): Test this color!

            red, green, blue = color
            return cls(red, green, blue)

        elif isinstance(color, str):
            color = color.strip()
            if color.startswith('#'):
                value = int(color[1:], 16)
                red = (value >> 16) & 0xFF
                green = (value >> 8) & 0xFF
                blue = value & 0xFF
                return cls(red, green, blue)
            elif color.startswith('0x'):
                value = int(color[2:], 16)
                red = (value >> 16) & 0xFF
                green = (value >> 8) & 0xFF
                blue = value & 0xFF
                return cls(red, green, blue)

            color = color.lower().replace(' ', '').replace('_', '').replace('-', '')
            if color in COLORS:
                value = COLORS[color]
                red = (value >> 16) & 0xFF
                green = (value >> 8) & 0xFF
                blue = value & 0xFF
                return cls(red, green, blue)
            else:
                raise SomeException  # OH NOW
        else:
            raise SomeException  # O NOW

        return Color.BLACK

    @property
    def r(self):
        return self._red

    @property
    def g(self):
        return self._green

    @property
    def b(self):
        return self._blue

    @property
    def rgb(self):
        return self.red, self.green, self.blue

    @property
    def hex(self):
        # TODO(sredmond): This is SO dumb. Fix this code!
        r = hex(self._red).replace('0x', '').zfill(2)
        g = hex(self._green).replace('0x', '').zfill(2)
        b = hex(self._blue).replace('0x', '').zfill(2)
        return '#{}{}{}'.format(r,g,b).upper()
        # return "#{:0>6}".format(hex(rgb)[2:].upper())

    @property
    def name(self):
        """Return a human-readable name for this color."""
        return "Nearly <color>"

    @classmethod
    def darken(cls, color):
        """Construct a new color at 2/3 of the current color's red, green, and blue values, rounded down."""
        if not isinstance(color, cls):
            color = cls.normalize(color)
        return cls(int(2 * color.r / 3), int(2 * color.g / 3), int(2 * color.b / 3))

    @classmethod
    def brighten(cls, color):
        if not isinstance(color, cls):
            color = cls.normalize(color)
        new_red = math.ceil(color.r + (255 - color.r) / 3)
        new_green = math.ceil(color.g + (255 - color.g) / 3)
        new_blue = math.ceil(color.b + (255 - color.b) / 3)
        return cls(new_red, new_green, new_blue)

    def __iter__(self):
        yield self.r
        yield self.g
        yield self.b


COLORS = {
    'aliceblue': 0xF0F8FF,
    'antiquewhite': 0xFAEBD7,
    'aqua': 0x00FFFF,
    'aquamarine': 0x7FFFD4,
    'azure': 0xF0FFFF,
    'beige': 0xF5F5DC,
    'bisque': 0xFFE4C4,
    'black': 0x000000,
    'blanchedalmond': 0xFFEBCD,
    'blue': 0x0000FF,
    'blueviolet': 0x8A2BE2,
    'brown': 0xA52A2A,
    'burlywood': 0xDEB887,
    'cadetblue': 0x5F9EA0,
    'chartreuse': 0x7FFF00,
    'chocolate': 0xD2691E,
    'coral': 0xFF7F50,
    'cornflowerblue': 0x6495ED,
    'cornsilk': 0xFFF8DC,
    'crimson': 0xDC143C,
    'cyan': 0x00FFFF,
    'darkblue': 0x00008B,
    'darkcyan': 0x008B8B,
    'darkgoldenrod': 0xB8860B,
    'darkgray': 0xA9A9A9,
    'darkgreen': 0x006400,
    'darkgrey': 0xA9A9A9,
    'darkkhaki': 0xBDB76B,
    'darkmagenta': 0x8B008B,
    'darkolivegreen': 0x556B2F,
    'darkorange': 0xFF8C00,
    'darkorchid': 0x9932CC,
    'darkred': 0x8B0000,
    'darksage': 0x598556,
    'darksalmon': 0xE9967A,
    'darkseagreen': 0x8FBC8F,
    'darkslateblue': 0x483D8B,
    'darkslategray': 0x2F4F4F,
    'darkslategrey': 0x2F4F4F,
    'darkturquoise': 0x00CED1,
    'darkviolet': 0x9400D3,
    'deeppink': 0xFF1493,
    'deepskyblue': 0x00BFFF,
    'dimgray': 0x696969,
    'dimgrey': 0x696969,
    'dodgerblue': 0x1E90FF,
    'firebrick': 0xB22222,
    'floralwhite': 0xFFFAF0,
    'forestgreen': 0x228B22,
    'fuchsia': 0xFF00FF,
    'gainsboro': 0xDCDCDC,
    'ghostwhite': 0xF8F8FF,
    'gold': 0xFFD700,
    'goldenrod': 0xDAA520,
    'gray': 0x808080,
    'green': 0x008000,
    'greenyellow': 0xADFF2F,
    'grey': 0x808080,
    'honeydew': 0xF0FFF0,
    'hotpink': 0xFF69B4,
    'indianred': 0xCD5C5C,
    'indigo': 0x4B0082,
    'ivory': 0xFFFFF0,
    'khaki': 0xF0E68C,
    'lavender': 0xE6E6FA,
    'lavenderblush': 0xFFF0F5,
    'lawngreen': 0x7CFC00,
    'lemonchiffon': 0xFFFACD,
    'lightblue': 0xADD8E6,
    'lightcoral': 0xF08080,
    'lightcyan': 0xE0FFFF,
    'lightgoldenrodyellow': 0xFAFAD2,
    'lightgray': 0xD3D3D3,
    'lightgreen': 0x90EE90,
    'lightgrey': 0xD3D3D3,
    'lightpink': 0xFFB6C1,
    'lightsage': 0xBCECAC,
    'lightsalmon': 0xFFA07A,
    'lightseagreen': 0x20B2AA,
    'lightskyblue': 0x87CEFA,
    'lightslategray': 0x778899,
    'lightslategrey': 0x778899,
    'lightsteelblue': 0xB0C4DE,
    'lightyellow': 0xFFFFE0,
    'lime': 0x00FF00,
    'limegreen': 0x32CD32,
    'linen': 0xFAF0E6,
    'magenta': 0xFF00FF,
    'maroon': 0x800000,
    'mediumaquamarine': 0x66CDAA,
    'mediumblue': 0x0000CD,
    'mediumorchid': 0xBA55D3,
    'mediumpurple': 0x9370DB,
    'mediumseagreen': 0x3CB371,
    'mediumslateblue': 0x7B68EE,
    'mediumspringgreen': 0x00FA9A,
    'mediumturquoise': 0x48D1CC,
    'mediumvioletred': 0xC71585,
    'midnightblue': 0x191970,
    'mintcream': 0xF5FFFA,
    'mistyrose': 0xFFE4E1,
    'moccasin': 0xFFE4B5,
    'navajowhite': 0xFFDEAD,
    'navy': 0x000080,
    'oldlace': 0xFDF5E6,
    'olive': 0x808000,
    'olivedrab': 0x6B8E23,
    'orange': 0xFFA500,
    'orangered': 0xFF4500,
    'orchid': 0xDA70D6,
    'palegoldenrod': 0xEEE8AA,
    'palegreen': 0x98FB98,
    'paleturquoise': 0xAFEEEE,
    'palevioletred': 0xDB7093,
    'papayawhip': 0xFFEFD5,
    'peachpuff': 0xFFDAB9,
    'peru': 0xCD853F,
    'pink': 0xFFC0CB,
    'plum': 0xDDA0DD,
    'powderblue': 0xB0E0E6,
    'purple': 0x800080,
    'red': 0xFF0000,
    'rosybrown': 0xBC8F8F,
    'royalblue': 0x4169E1,
    'saddlebrown': 0x8B4513,
    'sage': 0x87AE73,
    'salmon': 0xFA8072,
    'sandybrown': 0xFAA460,
    'seagreen': 0x2E8B57,
    'seashell': 0xFFF5EE,
    'sienna': 0xA0522D,
    'silver': 0xC0C0C0,
    'skyblue': 0x87CEEB,
    'slateblue': 0x6A5ACD,
    'slategray': 0x708090,
    'slategrey': 0x708090,
    'snow': 0xFFFAFA,
    'springgreen': 0x00FF7F,
    'steelblue': 0x4682B4,
    'tan': 0xD2B48C,
    'teal': 0x008080,
    'thistle': 0xD8BFD8,
    'tomato': 0xFF6347,
    'turquoise': 0x40E0D0,
    'violet': 0xEE82EE,
    'wheat': 0xF5DEB3,
    'white': 0xFFFFFF,
    'whitesmoke': 0xF5F5F5,
    'yellow': 0xFFFF00,
    'yellowgreen': 0x9ACD32
}

# Overrides from the Stanford ACM libraries, whose color
# definitions may be slightly different.
COLORS.update({
    "black": 0x000000,
    "darkgray": 0x595959,
    "darkgrey": 0x595959,
    "gray": 0x999999,
    "grey": 0x999999,
    "lightgray": 0xBFBFBF,
    "lightgrey": 0xBFBFBF,
    "orange": 0xFFC800,
    "pink": 0xFFAFAF,
})
